#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import posixpath
import numpy as np

from tilupy.utils import format_path_linux

import tilupy.notations
import tilupy.raster


README_PARAM_MATCH = dict(tmax="tmax", CFL="cflhyp", h_min="eps0", dt_im_output="dt_im")
"""Dictionary of correspondence between parameters in read_me file and param file."""

SHALTOP_LAW_ID = dict(
    No_Friction=1, Herschel_Bulkley=61, Voellmy=8, Bingham=6, Coulomb=1, Coulomb_muI=7
)
"""Dictionary of correspondence between the name of rheological laws and the corresponding ID in SHALTOP.

Correspondence :

    - No_Friction = 1
    - Coulomb = 1
    - Coulomb_muI = 7
    - Voellmy = 8
    - Bingham = 6
    - Herschel_Bulkley = 61
"""


def raster_to_shaltop_txtfile(
    file_in: str, file_out: str, folder_out: str = None
) -> dict:
    """Convert a raster file to a Shaltop ASCII text format.

    Reads a raster (formats readable by :func:`tilupy.raster.read_raster`) and saves it as a
    ASCII text file with values flattened column-wise and rows flipped vertically.

    Parameters
    ----------
    file_in : str
        Path to the input raster file.
    file_out : str
        Name of the output ASCII text file.
    folder_out : str, optional
        Directory where the output file will be saved. If None, `file_out`
        is used as-is, by default None.

    Returns
    -------
    dict
        Dictionary containing grid metadata:

            - 'x0': X coordinate of the first column.
            - 'y0': Y coordinate of the first row.
            - 'dx': Grid spacing along X.
            - 'dy': Grid spacing along Y.
            - 'nx': Number of columns.
            - 'ny': Number of rows.
    """
    if folder_out is not None:
        file_out = os.path.join(folder_out, file_out)

    x, y, rast = tilupy.raster.read_raster(file_in)
    np.savetxt(file_out, np.reshape(np.flip(rast, axis=0), (rast.size, 1)), fmt="%.12G")

    res = dict(x0=x[0], y0=y[0], dx=x[1] - x[0], dy=y[1] - y[0], nx=len(x), ny=len(y))

    return res


def write_params_file(
    params: dict, directory: str = None, file_name: str = "params.txt"
) -> None:
    """Write a dictionary of parameters to a text file.

    Each key-value pair in the dictionary is written on a separate line.

    Parameters
    ----------
    params : dict
        Dictionary of parameter names and values. Values can be int, float,
        or str.
    directory : str, optional
        Directory where the parameter file will be written. Default is the
        current working directory.
    file_name : str, optional
        Name of the parameter file. Default is "params.txt".
    """
    if directory is None:
        directory = os.getcwd()

    with open(os.path.join(directory, file_name), "w") as file_params:
        for name in params:
            val = params[name]
            if (
                isinstance(val, int)
                or isinstance(val, np.int64)
                or isinstance(val, np.int32)
            ):
                file_params.write("{:s} {:d}\n".format(name, val))

            if (
                isinstance(val, float)
                or isinstance(val, np.float64)
                or isinstance(val, np.float32)
            ):
                file_params.write("{:s} {:.8G}\n".format(name, val))

            if isinstance(val, str):
                file_params.write("{:s} {:s}\n".format(name, val))


def write_simu(
    raster_topo: str,
    raster_mass: str,
    tmax: float,
    dt_im: float,
    rheology_type: str,
    rheology_params: dict,
    folder_out: str = None,
) -> None:
    """
    Prepares the input files required for a SHALTOP simulation and saves them in a dedicated folder.

    Parameters
    ----------
    raster_topo : str
        Name of the ASCII topography file.
    raster_mass : str
        Name of the ASCII initial mass file.
    tmax : float
        Maximum simulation time.
    dt_im : float
        Output image interval (in time steps).
    rheology_type : str
        Rheology to use for the simulation.
    rheology_params : dict
        Necessary parameters for the rheology. For instance:

            - delta1
            - ksi
            - tau_density
            etc.
    folder_out : str, optional
        Output folder where simulation inputs will be saved.

    Raises
    ------
    ValueError
        If the rheology is wrong.
    """
    if folder_out is None:
        folder_out = "."

    # output_file = os.path.join(folder_out, "shaltop")

    os.makedirs(folder_out, exist_ok=True)

    x, y, z = tilupy.raster.read_raster(raster_topo)
    raster_to_shaltop_txtfile(raster_topo, os.path.join(folder_out, "z.d"))
    raster_to_shaltop_txtfile(raster_mass, os.path.join(folder_out, "m.d"))
    folder_output = "data2"
    os.makedirs(os.path.join(folder_out, folder_output), exist_ok=True)

    if rheology_type not in SHALTOP_LAW_ID:
        raise ValueError(f"Wrong law, choose in: {SHALTOP_LAW_ID}")

    params = dict(
        nx=len(x),
        ny=len(y),
        per=x[-1],
        pery=y[-1],
        tmax=tmax,
        dt_im=dt_im,
        initz=0,
        file_z_init="z.d",
        ipr=0,
        file_m_init="m.d",
        folder_output="data2",
        icomp=SHALTOP_LAW_ID[rheology_type],
        **rheology_params,
    )
    write_params_file(params, directory=folder_out, file_name="params.txt")


def write_job_files(
    dirs: list[str],
    param_files: list[str],
    file_job: str,
    job_name: str,
    max_time_hours: int = 24,
    ncores_per_node: int = 6,
    partitions: str = "cpuall,data,datanew",
    shaltop_file: str = "shaltop",
    folder_conf_in_job: str = None,
    replace_path: list = None,
    number_conf_file: bool = True,
) -> None:
    """
    Write job/conf files for slurm jobs. The conf contains all the commands
    needed to run each simulation (one command per simulation).

    Parameters
    ----------
    dirs : list[str]
        list of paths where simus will be run.
    param_files : list[str]
        list of shaltop parameter files.
    file_job : str
        name of job file called by sbatch.
    job_name : str
        name of conf file used by file_job.
    max_time_hours : int, optional
        Maximum job duration in hours before stop. The default is 24.
    ncores_per_node : int, optional
        Number of cores per nodes. Used to know the number of nodes required
        for the job. The default is 6.
    partitions : str, optional
        Names of partitions on which jobs can be launched.
        The default is "cpuall,data,datanew".
    shaltop_file : str, optional
        Bash command used to call shaltop. Can be a path.
        The default is "shaltop".
    folder_conf_in_job : str, optional
        Folder where the conf file is located. The default is the folder
        path of file_job.
    replace_path : list, optional
        replace replace_path[0] by replace_path[1] for every path in dir. This
        is used if simulations are prepared and run on two different machines
        (e.g. laptop and cluster).
        The default is None.
    number_conf_file : bool, optional
        If True, add a number in front of each line of the conf file. Required
        to identify slurm jobs.
        The default is True.
    """
    ntasks = len(dirs)
    nnodes = int(np.ceil(ntasks / ncores_per_node))

    if folder_conf_in_job is None:
        folder_conf_in_job = os.path.dirname(file_job)
        if folder_conf_in_job == "":
            folder_conf_in_job = "."

    with open(file_job + ".conf", "w", newline="\n") as conf_file:
        if number_conf_file:
            line = "{:d} {:s} {:s} {:s}\n"
        else:
            line = "{:s} {:s} {:s}\n"
        for i in range(ntasks):
            if replace_path is not None:
                folder = dirs[i].replace(replace_path[0], replace_path[1])
                param_file = param_files[i].replace(replace_path[0], replace_path[1])
            else:
                folder = dirs[i]
                param_file = param_files[i]
            folder = format_path_linux(folder)
            param_file = format_path_linux(param_file)
            if number_conf_file:
                line2 = line.format(i, shaltop_file, folder, param_file)
            else:
                line2 = line.format(shaltop_file, folder, param_file)
            conf_file.write(line2)

    n_hours = np.floor(max_time_hours)
    n_min = (max_time_hours - n_hours) * 60
    str_time = "{:02.0f}:{:02.0f}:00\n".format(n_hours, n_min)

    basename = os.path.basename(file_job)
    path_conf_in_job = posixpath.join(folder_conf_in_job, basename + ".conf")

    with open(file_job + ".job", "w", newline="\n") as job_file:
        job_file.write("#!/bin/sh\n")
        job_file.write("#SBATCH -J multijob\n")
        job_file.write("#SBATCH --job-name={:s}\n".format(job_name))
        job_file.write("#SBATCH --output={:s}%j.out\n".format(job_name))
        job_file.write("#SBATCH --partition " + partitions + "\n")
        job_file.write("#SBATCH --nodes={:d}".format(nnodes) + "\n")
        job_file.write("#SBATCH --ntasks={:d}".format(ntasks) + "\n")
        job_file.write("#SBATCH --time={:s}\n".format(str_time))
        job_file.write("\n")
        job_file.write("module purge\n")
        job_file.write("module load slurm\n")
        job_file.write("\n")
        line = "srun -n {:d} -l --multi-prog {:s}"
        job_file.write(line.format(ntasks, path_conf_in_job))


def make_simus(
    law: str, rheol_params: dict, folder_data: str, folder_out: str, readme_file: str
) -> None:
    """Write shaltop initial file for simple slope test case

    Reads topography and initial mass files in ASCII format,
    writes them in Shaltop-compatible format, prepares simulation parameters
    based on a README file and user-provided values,
    and generates a shell script to run the simulations.

    Parameters
    ----------
    law : str
        Name of the rheological law to use (must match a key in :data:`SHALTOP_LAW_ID`).
    rheol_params : dict of list
        Dictionary of rheology parameters. Each key corresponds to a parameter
        name and its value is a list of parameter values to simulate.
    folder_data : str
        Path to the folder containing input data files "topo.asc" and "mass.asc".
    folder_out : str
        Path to the folder where output simulation folders and Shaltop files
        will be created.
    readme_file : str
        Path to the README file containing simulation parameters and metadata.
    """
    # Get topography and initial mass, and write them in Shaltop format
    zfile = os.path.join(folder_data, "topo.asc")
    mfile = os.path.join(folder_data, "mass.asc")
    x, y, z, dx = tilupy.raster.read_ascii(zfile)
    _, _, m, _ = tilupy.raster.read_ascii(mfile)
    np.savetxt(os.path.join(folder_out, "z.d"), z.T.flatten())
    np.savetxt(os.path.join(folder_out, "m.d"), m.T.flatten())

    # Get simulation parameters from README.txt and raster .asc files
    params = tilupy.notations.readme_to_params(readme_file, README_PARAM_MATCH)
    params["nx"] = len(x)
    params["ny"] = len(y)
    params["per"] = dx * len(x)
    params["pery"] = dx * len(y)
    params["file_m_init"] = "../m.d"
    params["file_z_init"] = "../z.d"

    # Folder for rheological law, and set params accordingly
    folder_law = os.path.join(folder_out, law)
    params["icomp"] = SHALTOP_LAW_ID[law]

    param_names = [param for param in rheol_params]

    texts = tilupy.notations.make_rheol_string(rheol_params, law)

    # Run shaltop file
    run_shaltop_file = os.path.join(folder_law, "run_shaltop.sh")
    file_txt = ""

    for i in range(len(rheol_params[param_names[0]])):
        simu_text = texts[i]
        for param_name in param_names:
            params[param_name] = rheol_params[param_name][i]
        params["folder_output"] = simu_text
        folder_results = os.path.join(folder_law, simu_text)
        os.makedirs(folder_results, exist_ok=True)
        with open(os.path.join(folder_results, ".gitignore"), "w") as fid:
            fid.write("# Ignore everything in this directory")
            fid.write("*")
            fid.write("# Except this file")
            fid.write("!.gitignore")

        write_params_file(params, directory=folder_law, file_name=simu_text + ".txt")
        file_txt += "start_time=`date +%s`\n"
        file_txt += 'shaltop "" ' + simu_text + ".txt\n"
        file_txt += "end_time=`date +%s`\n"
        file_txt += "elapsed_time=$(($end_time - $start_time))\n"
        file_txt += 'string_time="${start_time} ' + simu_text + ' ${elapsed_time}"\n'
        file_txt += "echo ${string_time} >> simulation_duration.txt\n\n"

    with open(run_shaltop_file, "w") as fid:
        fid.write(file_txt)



# -*- coding: utf-8 -*-
"""
Created on Fri Aug  4 12:45:15 2023

@author: peruzzetto
"""

import numpy as np

import tilupy.make_topo


def test_topo_gray():
    x, y, z = tilupy.make_topo.gray99()

    assert (np.abs(y[-1] - 0.6) < 1e-6) & (len(x) == 361)



import pytest
import os

import tilupy.models.saval2D.read as saread
import tilupy.read as tiread


@pytest.fixture
def simu_data():
    res = dict(
        folder=os.path.join(os.path.dirname(__file__), "data", "gray99", "saval2D"),
        raster_name="mntsimulation",
    )
    return res


@pytest.mark.parametrize(
    "args, expected",
    [
        ("h", (3, tiread.TemporalResults2D)),
        ("u", (3, tiread.TemporalResults2D)),
        ("hu", (3, tiread.TemporalResults2D)),
        ("hu2", (3, tiread.TemporalResults2D)),
        ("ux", (3, tiread.TemporalResults2D)),
        ("uy", (3, tiread.TemporalResults2D)),
        ("hvert", (3, tiread.TemporalResults2D)),
    ],
)
def test_extract_output(simu_data, args, expected):
    res = saread.Results(
        folder=simu_data["folder"], raster_topo=simu_data["raster_name"]
    )
    output = res._extract_output(args)

    assert output.d.ndim == expected[0]
    assert isinstance(output, expected[1])


def test_extract_invalid_output(simu_data):
    res = saread.Results(
        folder=simu_data["folder"], raster_topo=simu_data["raster_name"]
    )
    output = res._extract_output("xx")

    assert isinstance(output, tiread.AbstractResults)



import pytest
import os

import numpy

import tilupy.models.shaltop.read as sharead
import tilupy.read as tiread


@pytest.fixture
def simu_data():
    res = dict(
        folder=os.path.join(os.path.dirname(__file__), "data", "gray99", "shaltop"),
        param_file="params.txt",
    )
    return res


def test_read_param(simu_data):
    param_file = os.path.join(simu_data["folder"], simu_data["param_file"])
    di = sharead.read_params(param_file)
    assert list(di.keys()) == [
        "nx",
        "ny",
        "per",
        "pery",
        "tmax",
        "dt_im",
        "initz",
        "file_z_init",
        "ipr",
        "file_m_init",
        "folder_output",
        "icomp",
        "tau_density",
        "K_tau",
    ]


def test_read_bin(simu_data):
    bin_file = os.path.join(simu_data["folder"], "data2", "ekmax.bin")
    data = sharead.read_file_bin(bin_file, 236, 68)
    assert isinstance(data, numpy.ndarray)


def test_read_init(simu_data):
    init_file = os.path.join(simu_data["folder"], "z.d")
    data = sharead.read_file_init(init_file, 236, 68)
    assert isinstance(data, numpy.ndarray)


@pytest.mark.parametrize(
    "args, expected",
    [
        ("h", (3, tiread.TemporalResults2D)),
        ("ux", (3, tiread.TemporalResults2D)),
        ("uy", (3, tiread.TemporalResults2D)),
        ("hvert", (3, tiread.TemporalResults2D)),
        ("u", (3, tiread.TemporalResults2D)),
        ("hu", (3, tiread.TemporalResults2D)),
        ("hu2", (3, tiread.TemporalResults2D)),
        ("ek", (1, tiread.TemporalResults0D)),
        ("ep", (1, tiread.TemporalResults0D)),
        ("etot", (1, tiread.TemporalResults0D)),
        ("faccx", (3, tiread.TemporalResults2D)),
        ("fcurvx", (3, tiread.TemporalResults2D)),
        ("ffricx", (3, tiread.TemporalResults2D)),
        ("fgravx", (3, tiread.TemporalResults2D)),
        ("finertx", (3, tiread.TemporalResults2D)),
        ("fpressionx", (3, tiread.TemporalResults2D)),
        ("shearx", (3, tiread.TemporalResults2D)),
        ("normalx", (3, tiread.TemporalResults2D)),
        ("pbottom", (3, tiread.TemporalResults2D)),
        ("pcorrdt", (3, tiread.TemporalResults2D)),
        ("pcorrdiv", (3, tiread.TemporalResults2D)),
    ],
)
def test_extract_output(simu_data, args, expected):
    res = sharead.Results(simu_data["folder"], file_params=simu_data["param_file"])
    output = res._extract_output(args)

    assert output.d.ndim == expected[0]
    assert isinstance(output, expected[1])


def test_extract_invalid_output(simu_data):
    res = sharead.Results(simu_data["folder"], file_params=simu_data["param_file"])
    output = res._extract_output("xx")

    assert isinstance(output, tiread.AbstractResults)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jun  1 15:27:36 2021

@author: peruzzetto
"""

import os

import numpy as np
import tilupy.read

# Dictionnary with results names lookup table, to match code output names
LOOKUP_NAMES = dict(
    h="hflow",
    u="vflow",
    ux="vflowx",
    uy="vflowy",
    hu="momentum",
    hu2="ek",
    vol="vol",
    ekint="ekint",
)

# Classify results
STATES_OUTPUT = ["h", "ux", "uy"]
forces = ["facc", "fcurv", "ffric", "fgrav", "fpression"]
FORCES_OUTPUT = []
for axis in ["x", "y"]:
    for f in forces:
        FORCES_OUTPUT.append(f + axis)


def read_params(file):
    """
    Read simulation parameters from file.

    Parameters
    ----------
    file : str
        Parameters file.

    Returns
    -------
    params : dict
        Dictionnary with parameters

    """
    params = dict()
    if not file.endswith(".txt"):
        prefix = file.sep("/")[-1].sep("_")[0]
        file = os.path.join(file, prefix + "_files", prefix + "_documentation.txt")

    with open(file, "r") as f:
        for line in f:
            (key, val) = line.split("\t")
            try:
                params[key] = float(val)
            except ValueError:
                if val == "TRUE":
                    params[key] = True
                if val == "FALSE":
                    params[key] = False
                else:
                    params[key] = val

    return params


def read_ascii(file):
    """
    Read ascii grid file to numpy ndarray.

    Parameters
    ----------
    file : TYPE
        DESCRIPTION.

    Returns
    -------
    None.

    """
    dem = np.loadtxt(file, skiprows=6)
    # dem = np.flip(dem, axis=0).T
    grid = {}
    with open(file, "r") as fid:
        for i in range(6):
            tmp = fid.readline().split()
            grid[tmp[0]] = float(tmp[1])
    try:
        x0 = grid["xllcenter"]
        y0 = grid["yllcenter"]
    except KeyError:
        x0 = grid["xllcorner"]
        y0 = grid["yllcorner"]
    nx = int(grid["ncols"])
    ny = int(grid["nrows"])
    dx = dy = grid["cellsize"]
    x = np.linspace(x0, x0 + (nx - 1) * dx, nx)
    y = np.linspace(y0, y0 + (ny - 1) * dy, ny)

    return x, y, dem


def read_asciis(file_prefix, folder=None, ind=None, nodigit=False):
    """
    Read mutiple ascii grid file matching prefix.

    Parameters
    ----------
    file : TYPE
        DESCRIPTION.
    ind : TYPE, optional
        DESCRIPTION. The default is None.

    Returns
    -------
    None.

    """
    if folder is None:
        folder = os.getcwd()

    files_tmp = os.listdir(folder)
    files = []
    for ff in files_tmp:
        # test is file begins with file_prefix
        if ff.startswith(file_prefix) and ff.endswith(".asc"):
            letter = ff.split(file_prefix)[1][0]
            # After file_prefix, character must be either '.' or a digit
            if nodigit:
                if letter == ".":
                    files.append(ff)
            else:
                if letter.isdigit() or letter == ".":
                    files.append(ff)

    files.sort()

    if ind is not None:
        if ind == "final":
            files = [files[-1]]
        elif ind == "initial":
            files = [files[0]]
        else:
            files = [ff for ff in files if int(ff[-8:-4]) in ind]

    files = [os.path.join(folder, ff) for ff in files]

    dem = np.loadtxt(files[0], skiprows=6)
    if len(files) > 1:
        dem = dem[:, :, np.newaxis]
    for ff in files[1:]:
        _, _, dem2 = np.loadtxt(ff, skiprows=6)
        dem = np.concatenate((dem, dem2[:, :, np.newaxis]), axis=2)

    dem = np.squeeze(dem)

    return dem


class Results(tilupy.read.Results):
    """Results of shaltop simulations."""

    def __init__(self, text_rheol=None, folder_base=None, **varargs):
        """
        Init simulation results.

        Parameters
        ----------
        file_params : str
            File where simulation parameters will be read

        """
        self.code = "ravaflow"
        self.prefix = text_rheol
        self.folder_base = folder_base
        file_params = os.path.join(
            folder_base,
            self.prefix + "_results",
            self.prefix + "_files",
            self.prefix + "_documentation.txt",
        )
        self.params = read_params(file_params)

        self.folder_ascii = os.path.join(
            folder_base, self.prefix + "_results", self.prefix + "_ascii"
        )
        self.folder_files = os.path.join(
            folder_base, self.prefix + "_results", self.prefix + "_files"
        )

        self.set_axes()

        # Get time of outputs
        with open(os.path.join(self.folder_files, self.prefix + "_summary.txt")) as fid:
            self.tim = []
            for line in fid:
                res = line.split()
                if len(res) > 0:
                    if res[0].isdigit():
                        self.tim.append(float(res[4]))

    def set_axes(self, x=None, y=None, **varargs):
        """Set x and y axes."""
        dd = self.params["Cell size"]
        x = np.arange(
            self.params["Western boundary"] + dd / 2,
            self.params["Eastern boundary"],
            dd,
        )
        y = np.arange(
            self.params["Southern boundary"] + dd / 2,
            self.params["Northern boundary"],
            dd,
        )
        self.x = x
        self.y = y
        self.nx = len(x)
        self.ny = len(y)

    def set_zinit(self, zinit=None):
        """Set zinit, initial topography."""
        path_zinit = os.path.join(self.folder_ascii, self.prefix + "_elev.asc")
        self.zinit = np.loadtxt(path_zinit, skiprows=6)

    def get_temporal_output(self, name, d=None, t=None, **varargs):
        """
        Read 2D time dependent simulation results.

        Parameters
        ----------
        name : str
            Name of output.
        d : ndarray, optional
            Data to be read. If None, will be computed.
            The default is None.
        t : ndarray, optional
            Time of results snapshots (1D array), matching last dimension of d.
            If None, will be computed. The default is None.
        **varargs : TYPE
            DESCRIPTION.

        """
        # Read thicknesses or velocity components
        file_prefix = self.prefix + "_" + LOOKUP_NAMES[name]
        d = read_asciis(file_prefix, folder=self.folder_ascii)

        return tilupy.read.TemporalResults(name, d, self.tim)

    def get_static_output(self, name, stat, d=None, from_file=False, **varargs):
        """
        Read 2D time dependent simulation results.

        Parameters
        ----------
        name : str
            Name of output.
        d : ndarray, optional
            Data to be read. Last dimension is for time.
            If None, will be computed.
            The default is None.
        **varargs : TYPE
            DESCRIPTION.

        """
        if stat in ["initial", "final"]:
            file_prefix = self.prefix + "_" + LOOKUP_NAMES[name]
            d = read_asciis(file_prefix, folder=self.folder_ascii, ind=stat)
        elif stat == "max" and from_file:
            file = "{:s}_{:s}_{:s}.asc".format(self.prefix, LOOKUP_NAMES[name], stat)
            file = os.path.join(self.folder_ascii, file)
            d = np.loadtxt(file, skiprows=6)
        else:
            data = self.get_temporal_output(name)
            d = data.get_temporal_stat(stat).d

        return tilupy.read.StaticResults(name + stat, d)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
import matplotlib
import matplotlib.pyplot as plt
import seaborn as sns
import pytopomap.plot as pyplt
import pandas as pd

import tilupy.download_data


def plot_shotgather(
    x: np.ndarray,
    t: np.ndarray,
    data: np.ndarray,
    xlabel: str = "X (m)",
    ylabel: str = "Time (s)",
    **kwargs,
) -> matplotlib.axes._axes.Axes:
    """Plot shotgather image.

    Plot shotgather like image, with vertical axis as time and horizontal axis
    and spatial dimension. This is a simple call to plot_shotgather, but
    input data is transposed because in tilupy the last axis is time by
    convention.

    Parameters
    ----------
    x : numpy.ndarray
        Spatial coordinates, size NX.
    t : numpy.ndarray
        Time array (assumed in seconds), size NT.
    data : numpy.ndarray
        NX*NT array of data to be plotted.
    xlabel : string, optional
        Label for x-axis, by default "X (m)".
    ylabel : string, optional
        Label for x-axis, by default "Time (s)".
    **kwargs : dict, optional
        parameters passed on to :func:`pytopomap.plot.plot_imshow`.

    Returns
    -------
    matplotlib.axes._axes.Axes
        Axes instance where data is plotted
    """
    if "aspect" not in kwargs:
        kwargs["aspect"] = "auto"
    axe = pyplt.plot_imshow(x, t[::-1], data.T, **kwargs)
    axe.set_adjustable("box")
    axe.set_ylabel(ylabel)
    axe.set_xlabel(xlabel)

    return axe


def plot_heatmaps(
    df,
    values,
    index,
    columns,
    axs=None,
    aggfunc="mean",
    figsize=None,
    ncols=3,
    heatmap_kws=None,
    notations=None,
    best_values=None,
    plot_best_value="point",
    text_kwargs=None,
) -> matplotlib.figure.Figure:
    """Plot one or several heatmaps from a pandas DataFrame.

    Each heatmap is created by pivoting the DataFrame with the given
    `index`, `columns`, and a variable from `values`.

    Parameters
    ----------
    df : pandas.DataFrame
        Input DataFrame containing the data.
    values : list[str]
        Column names in :data:`df` to plot as separate heatmaps.
    index : str
        Column name to use as rows of the pivot table.
    columns : str
        Column name to use as columns of the pivot table.
    axes : matplotlib.Axes or list(matplotlib.Axes), optional
        Axe or list of Axes where results will be plotted. If None,
        they are created iteratively. By default, None.
    aggfunc : str or callable, optional
        Aggregation function applied when multiple values exist for
        a given (index, column) pair. By default "mean".
    figsize : tuple of float, optional
        Size of the matplotlib figure, by default None.
    ncols : int, optional
        Maximum number of heatmaps per row, by default 3.
    heatmap_kws : dict or dict[dict], optional
        Keyword arguments passed to :data:`seaborn.heatmap`.
        If dict of dict, keys must match the values in :data:`values`.
    notations : dict, optional
        Mapping from variable names to readable labels
        (used for axis and colorbar labels).
    best_values : dict, optional
        Mapping from variable names to selection criterion:
        "min", "min_abs", or "max".
    plot_best_value : {"point", "text"}, optional
        How to highlight best values:
        - "point" : mark with circles
        - "text" : display numeric values
        By default "point".
    text_kwargs : dict, optional
        Keyword arguments passed to :data:`matplotlib.axes.Axes.text` when
        annotating best values. Only used if :data:`plot_best_value="text"`.

    Returns
    -------
    matplotlib.figure.Figure
        The matplotlib Figure containing the heatmaps.
    """
    nplots = len(values)
    add_axs = axs is None
    if add_axs:
        ncols = min(nplots, ncols)
        nrows = int(np.ceil(nplots / ncols))
        fig = plt.figure(figsize=figsize)
        axes = []
    else:
        axes = axs
        fig = axs.flat[0].figure

    for i in range(nplots):
        if add_axs:
            axe = fig.add_subplot(nrows, ncols, i + 1)
            axes.append(axe)
        else:
            axe = axes.flat[i]
        data = df.pivot_table(
            index=index, columns=columns, values=values[i], aggfunc=aggfunc
        ).astype(float)
        if heatmap_kws is None:
            kws = dict()
        elif isinstance(heatmap_kws, dict):
            if values[i] in heatmap_kws:
                kws = heatmap_kws[values[i]].copy()
            else:
                kws = heatmap_kws.copy()

        if "cmap" not in kws:
            minval = data.min().min()
            maxval = data.max().max()
            if minval * maxval < 0:
                val = max(np.abs(minval), maxval)
                kws["cmap"] = "seismic"
                kws["vmin"] = -val
                kws["vmax"] = val

        if "cbar_kws" not in kws:
            kws["cbar_kws"] = dict(pad=0.03)

        if notations is None:
            kws["cbar_kws"]["label"] = values[i]
        else:
            if values[i] in notations:
                kws["cbar_kws"]["label"] = notations[values[i]]
            else:
                kws["cbar_kws"]["label"] = values[i]

        sns.heatmap(data, ax=axe, **kws)

        if best_values is not None:
            best_value = best_values[values[i]]
            array = np.array(data)
            irow = np.arange(data.shape[0])

            if best_value == "min":
                ind = np.nanargmin(array, axis=1)
                i2 = np.nanargmin(array[irow, ind])
            if best_value == "min_abs":
                ind = np.nanargmin(np.abs(array), axis=1)
                i2 = np.nanargmin(np.abs(array[irow, ind]))
            elif best_value == "max":
                ind = np.nanargmax(array, axis=1)
                i2 = np.nanargmax(array[irow, ind])

            if plot_best_value == "point":
                axe.plot(
                    ind + 0.5,
                    irow + 0.5,
                    ls="",
                    marker="o",
                    mfc="w",
                    mec="k",
                    mew=0.5,
                    ms=5,
                )
                axe.plot(
                    ind[i2] + 0.5,
                    i2 + 0.5,
                    ls="",
                    marker="o",
                    mfc="w",
                    mec="k",
                    mew=0.8,
                    ms=9,
                )
            elif plot_best_value == "text":
                indx = list(ind)
                indx.pop(i2)
                indy = list(irow)
                indy.pop(i2)
                default_kwargs = dict(ha="center", va="center", fontsize=8)
                if text_kwargs is None:
                    text_kwargs = default_kwargs
                else:
                    text_kwargs = dict(default_kwargs, **text_kwargs)
                for i, j in zip(indx, indy):
                    axe.text(
                        i + 0.5, j + 0.5, "{:.2g}".format(array[j, i]), **text_kwargs
                    )
                text_kwargs2 = dict(text_kwargs, fontweight="bold")
                axe.text(
                    ind[i2] + 0.5,
                    i2 + 0.5,
                    "{:.2g}".format(array[i2, ind[i2]]),
                    **text_kwargs2,
                )

    if add_axs:
        axes = np.array(axes).reshape((nrows, ncols))
        for i in range(nrows):
            for j in range(1, ncols):
                axes[i, j].set_ylabel("")
                # axes[i, j].set_yticklabels([])

        for i in range(nrows - 1):
            for j in range(ncols):
                axes[i, j].set_xlabel("")
                # axes[i, j].set_xticklabels([])

    if notations is not None:
        if add_axs:
            for i in range(nrows):
                axes[i, 0].set_ylabel(notations[index])
            for j in range(ncols):
                axes[-1, j].set_xlabel(notations[columns])
        else:
            for ax in axes.flatten():
                ax.set_ylabel(notations[index])
                ax.set_xlabel(notations[columns])

    # fig.tight_layout()

    return fig, axes


def plot_shaltop_mus_calibrated(
    data: pd.DataFrame = None,
    publication_date: str = None,
    plot_lucas_law: bool = True,
    step_delta: float = 5,
    ax: matplotlib.axes._axes.Axes = None,
    figsize: tuple = None,
) -> matplotlib.axes._axes.Axes:
    """Plot friction coefficient calibrated with Shaltop

    Plot the friction coefficient calibrated with Shaltop as a function of landslide
    volume, with seaborn.scatterplot. The data used to create the plot is downloaded from zenodo if not given

    Parameters
    ----------
    data : pd.DataFrame
        pd.DataFrame containing the calibrated values. If not provided, the data is downloaded
        from zenodo.
    publication_date : str
        string YYYY-MM-DD with the date for the data version
    plot_lucas_law : bool
        Plot the lucas law mu=V**(-0.0774) with associated confidence interval
    step_delta : float
        Step between delta angles on the right axis. Default is 5.
    ax : matplotlib.axes._axes.Axes
        Axe instance where plot must be done. If None, an axe instance is created. Deafault is None.
    figsize : tuple
        If ax is None, size of the figure to be created in inches. Default is None.
    """

    if data is None:
        data, publication_date = tilupy.download_data.import_shaltop_mus_calibrated()
    # Default plot options, are updated with provided kwargs
    plot_kw = dict(data=data, x="volume (m3)", y="mus", hue="calibration data")

    if ax is None:
        _, ax = plt.subplots(figsize=figsize)
    ax.grid(axis="both", which="both", zorder=-1)
    sns.scatterplot(**plot_kw, ax=ax, zorder=1)
    ax.set_xscale("log")
    # ax.set_yscale('log')

    # Decorate
    ax.set_xlabel("Volume (m$^3$)")
    ax.set_ylabel("$\\mu_S$")

    if publication_date is not None:
        ax.set_title("10.5281/zenodo.18791118, " + publication_date)

    if plot_lucas_law:
        vmin, vmax = ax.get_xlim()
        v = np.linspace(np.log10(vmin), np.log10(vmax))
        v = 10**v
        mu_lucas = v ** (-0.0774)
        mu_inf = mu_lucas * 10 ** (0.0667 * 1.96)
        mu_sup = mu_lucas * 10 ** (-0.0667 * 1.96)
        ax.fill_between(
            v,
            mu_inf,
            mu_sup,
            edgecolor=[0, 0, 0, 0],
            facecolor=[0.5, 0.5, 0.5, 0.3],
            zorder=0.5,
        )
        ax.plot(v, mu_lucas, "-", color="grey", zorder=0.5)
        ax.set_xlim(vmin, vmax)

    # Add corresponding deltas values on the right y axis
    mu_min, mu_max = ax.get_ylim()
    delta_min = np.rad2deg(np.arctan(mu_min))
    delta_min = step_delta * np.ceil(delta_min / step_delta)
    delta_max = np.rad2deg(np.arctan(mu_max))
    delta_max = step_delta * np.floor(delta_max / step_delta)
    ax2 = ax.twinx()
    ax2.set_ylim([mu_min, mu_max])
    ax2.grid(False)
    deltas = np.arange(delta_min, delta_max + step_delta / 2, step_delta)
    mu_deltas = np.tan(np.deg2rad(deltas))
    ax2.set_yticks(mu_deltas)
    ax2.set_yticklabels(["{:.0f}".format(delta) for delta in deltas])
    ax2.set_ylabel("$\\delta$ (°)")

    return ax



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed May 26 11:03:56 2021

@author: peruzzetto
"""

import os
import numpy as np

import tilupy.notations

readme_param_match = dict(tmax="tmax", CFL="cflhyp", h_min="eps0", dt_im_output="dt_im")


def readme_to_params(folder_data: str) -> dict:
    """Read a README.txt file and convert it to a parameters dictionary.

    Each line of the README file should contain a key and a value separated
    by whitespace. The function parses the file and returns a dictionary
    mapping keys to their corresponding string values.

    Parameters
    ----------
    folder_data : str
        Path to the folder containing the README.txt file.

    Returns
    -------
    dict
        Dictionary where keys are parameter names (str) and values are
        their corresponding values (str) read from the file.
    """
    params = dict()
    with open(os.path.join(folder_data, "README.txt"), "r") as f:
        for line in f:
            (key, val) = line.split()
            params[key] = val
    return params


def write_params(params: dict) -> str:
    """Convert a parameters dictionary to a formatted string for RavaFlow input.

    Generates a single string in the form "key1=val1 key2=val2 ...",
    suitable for passing as command-line arguments to RavaFlow.

    Parameters
    ----------
    params : dict
        Dictionary of parameters, where keys and values are strings.

    Returns
    -------
    str
        Formatted string containing all parameters in "key=value" format,
        separated by spaces.
    """
    txt = ""
    for key in params:
        txt += "{:s}={:s} ".format(key, params[key])
    return txt[:-1]


def make_simus(
    law: str, rheol_params: dict, folder_files: str, folder_out: str, readme_file: str
) -> None:
    """Write ravaflow initial file for simple slope test case

    Reads simulation parameters from a README file and input
    ASCII rasters, sets topography and mass, prepares numerical and rheological
    parameters, and generates a shell script to execute simulations. Supports
    multiple combinations of rheology parameters.

    Parameters
    ----------
    law : str
        Rheological law to use, e.g., "coulomb" or "voellmy".
    rheol_params : dict of list
        Dictionary of rheology parameters, each key maps to a list of values
        defining multiple simulation runs.
    folder_files : str
        Path to folder containing input ASCII files "topo.asc" and "mass.asc".
    folder_out : str
        Path to the folder where output simulation folders and run scripts
        will be created.
    readme_file : str
        Path to the README.txt file containing simulation metadata and parameters.

    Returns
    -------
    None
    """
    # Parameters from README.txt file
    params_readme = tilupy.notations.readme_to_params(readme_file)
    # Parameters for simulation
    params = dict()
    # Directory where run script will be created
    folder_law = os.path.join(folder_out, law)
    os.makedirs(folder_law, exist_ok=True)

    # Prepare simulation parameters

    # Topography and initial mass
    params["elevation"] = "elev"
    params["hrelease"] = "minit"

    # Control parameters
    params["controls"] = "0,0,0,0,0,0"

    # Thresholds parameters
    # (min thicknees for output, min flow knetic energy for output,
    # min flow pressure for output, min thickness in simulation)
    h0 = np.float(params_readme["h_min"])
    params["thresholds"] = "{:.2E},{:.2E},{:.2E},{:.2E}".format(h0, h0, h0, h0)

    # CFL parameters
    # (CFL condition, initial time step when CFL not applicable)
    params["cfl"] = "{:.2E},0.0001".format(np.float(params_readme["CFL"]))

    # Output time parameters
    params["time"] = "{:.3f},{:.2f}".format(
        np.float(params_readme["dt_im_output"]), np.float(params_readme["tmax"])
    )

    # Simulation phases
    params["phases"] = "s"

    # Write bash run file
    file_txt = ""
    file_path = os.path.join(folder_law, "run.avaflow.sh")

    # Set region
    grid_params = dict()
    with open(os.path.join(folder_files, "topo.asc"), "r") as fid:
        for i in range(6):
            tmp = fid.readline().split()
            grid_params[tmp[0]] = float(tmp[1])
    e = grid_params["cellsize"] * grid_params["ncols"]
    n = grid_params["cellsize"] * grid_params["nrows"]
    tmp = "g.region s=0 w=0 e={:.2f} n={:.2f} res={:.4f}\n"
    file_txt += tmp.format(e, n, grid_params["cellsize"])
    file_txt += "g.region -s\n\n"

    # Read ascii files for mass and topography
    # Get relative path to topo and mass file, from the fodler
    # where simulations will be run. Find position of folder_files in
    # folder_out path. Simulations are run in folder_out/law/
    path = os.path.normpath(folder_out)
    ffs = path.split(os.sep)
    _, folder_topo = os.path.split(folder_files)
    n_up = len(ffs) - ffs.index(folder_topo)
    path_up = "../" * n_up
    zfile = path_up + "topo.asc"
    mfile = path_up + "mass.asc"
    file_txt += "r.in.gdal --overwrite input={:s} output=elev\n".format(zfile)
    file_txt += "r.in.gdal --overwrite input={:s} output=minit\n".format(mfile)

    # prefix for output folder
    prefixs = tilupy.notations.make_rheol_string(rheol_params, law)

    param_names = [param for param in rheol_params]

    txt_friction = "{:.2f},{:.2f},{:.2f}"

    for i in range(len(rheol_params[param_names[0]])):
        simu_prefix = prefixs[i]
        if law == "coulomb":
            delta1 = rheol_params["delta1"][i]
            params["friction"] = txt_friction.format(delta1, delta1, 0)

        if law == "voellmy":
            delta1 = rheol_params["delta1"][i]
            ksi = rheol_params["ksi"][i]
            params["friction"] = txt_friction.format(delta1, delta1, np.log10(ksi))

        params["prefix"] = simu_prefix

        file_txt += "\n"
        file_txt += "start_time=`date +%s`\n"
        file_txt += "r.avaflow -e " + write_params(params) + "\n"
        file_txt += "end_time=`date +%s`\n"
        file_txt += "elapsed_time=$(($end_time - $start_time))\n"
        file_txt += 'string_time="${start_time} ' + simu_prefix + ' ${elapsed_time}"\n'
        file_txt += "echo ${string_time} >> simulation_duration.txt\n\n"

    with open(file_path, "w") as fid:
        fid.write(file_txt)



_default_config = dict(language="english")


class Config(dict):
    def __init__(self, **kwargs):
        for key in kwargs:
            self[key] = kwargs[key]

    def update(self, **kwargs):
        for key in kwargs:
            self[key] = kwargs[key]


config = Config(**_default_config)


def set_config(**kwargs):
    config.update(**kwargs)

    if "language" in kwargs:
        import tilupy.notations

        tilupy.notations.set_labels()


# import tilupy.notations



# -*- coding: utf-8 -*-

import numpy as np


def calotte(
    x: np.ndarray,
    y: np.ndarray,
    z: np.ndarray,
    x0: float,
    y0: float,
    radius: float,
    norm_offset: float = 0,
    res_type: str = "projected_normal",
) -> np.ndarray:
    """Construct mass on topography as volume between sphere and topography.

    Parameters
    ----------
    x : numpy.ndarray
        X-axis array, with length nx.
    y : numpy.ndarray
        Y-axis array, with length ny.
    z : numpy.ndarray
        Array of altitudes, of size (ny, nx). z[0, 0] has coordinates
        (x[0], y[-1]).
    x0 : float
        X position of the calotte.
    y0 : float
        Y position of the calotte.
    radius : float
        Radius of the shpere.
    norm_offset : float, optional
        Downwards offset between the sphere center and the topography, in the
        direction normal to the topography, by default 0.
    res_type : string, optional
        Type of thickness output:

            - 'true_normal': Real thickness in the direction normal to the topography.
            - 'vertical': Thickness in the vertical direction.
            - 'projected_normal': Thickness normal to the topography is computed from the vertical
              thickness projected on the axe normal to the topography.

        The default is 'projected_normal'.

    Returns
    -------
    numpy.ndarray
        Array of mass height, in the direction normal to topography.
    """
    z = np.flip(z, axis=0).T

    xmesh, ymesh = np.meshgrid(x, y, indexing="ij")
    nx = len(x)
    ny = len(y)

    # Get altitude of mass center on topography
    i0 = np.unravel_index(np.argmin(np.abs(x - x0), axis=None), (nx,))
    j0 = np.unravel_index(np.argmin(np.abs(y - y0), axis=None), (ny,))
    z0 = z[i0, j0]
    # Topography gradient
    [Fx, Fy] = np.gradient(z, x, y, edge_order=2)
    Fz = np.ones((nx, ny))
    c = 1 / np.sqrt(1 + Fx**2 + Fy**2)
    Fx = -Fx * c
    Fy = -Fy * c
    Fz = Fz * c
    # Correct position from offset (shpere is moved downward,
    # perpendicular to topography)
    x0 = x0 - norm_offset * Fx[i0, j0]
    y0 = y0 - norm_offset * Fy[i0, j0]
    z0 = z0 - norm_offset * Fz[i0, j0]

    # Compute mass height only where relevant (ie around the mass center)
    dist_to_mass = (xmesh - x0) ** 2 + (ymesh - y0) ** 2
    ind = dist_to_mass <= radius**2

    B = 2 * (Fx * (xmesh - x0) + Fy * (ymesh - y0) + Fz * (z - z0))
    C = (xmesh - x0) ** 2 + (ymesh - y0) ** 2 + (z - z0) ** 2 - radius**2
    D = B**2 - 4 * C

    # Intersection between shpere and normal to the topography, solution of
    # t**2+B*t+C=0
    m = np.zeros((nx, ny))

    if res_type == "true_normal":
        # B = 2*(Fx[ind]*(xmesh[ind]-x0)+Fy[ind] *
        #        (ymesh[ind]-y0)+Fz[ind]*(z[ind]-z0))
        # C = (xmesh[ind]-x0)**2+(ymesh[ind]-y0)**2+(z[ind]-z0)**2-radius**2
        # D = B**2-4*C
        B = 2 * (Fx * (xmesh - x0) + Fy * (ymesh - y0) + Fz * (z - z0))
        C = (xmesh - x0) ** 2 + (ymesh - y0) ** 2 + (z - z0) ** 2 - radius**2
        D = B**2 - 4 * C
        ind = D > 0
        t1 = (-B - np.sqrt(D)) / 2
        t2 = (-B + np.sqrt(D)) / 2
        ind2 = t1 * t2 < 0
        m[ind2] = np.maximum(t1[ind2], t2[ind2])

    # Vertical thickness of calotte.
    if res_type in ["vertical", "projected_normal"]:
        zs = z0 + np.sqrt(radius**2 - (xmesh - x0) ** 2 - (ymesh - y0) ** 2)
        zi = z0 - np.sqrt(radius**2 - (xmesh - x0) ** 2 - (ymesh - y0) ** 2)
        ind = (z < zs) & (z > zi)
        m[ind] = zs[ind] - z[ind]
        if res_type == "projected_normal":
            m = m * c

    m = np.flip(m.T, axis=0)

    return m



# -*- coding: utf-8 -*-

import importlib


def write(
    model_name: str,
    raster_topo: str,
    raster_mass: str,
    tmax: float,
    dt_im: float,
    rheology_type: str,
    rheology_params: dict = None,
    folder_out: str = None,
    **kwargs,
):
    """
    Dynamically imports the corresponding initiation module from
    :data:`tilupy.models.<code>.initsimus` and use the corresponding :data:`write_simu()` function.

    Parameters
    ----------
    model_name : str
        Model to create simulation files.
    raster_topo : str, optional
        Path for an ASCII topography.
    raster_mass : str, optional
        Path for an ASCII initial mass.
    tmax : float
        Maximum simulation time.
    dt_im : float
        Output image interval (in time steps).
    rheology_type : str
        Rheology to use for the simulation.
    rheology_params : dict
        Parameters specific to the selected rheology.
    folder_out : str, optional
        Output folder where simulation inputs will be saved.
    **kwargs
        Additional arguments for specific models.
    """
    module = importlib.import_module("tilupy.models." + model_name + ".initsimus")

    module.write_simu(
        raster_topo=raster_topo,
        raster_mass=raster_mass,
        tmax=tmax,
        dt_im=dt_im,
        rheology_type=rheology_type,
        rheology_params=rheology_params,
        folder_out=folder_out,
        **kwargs,
    )



# -*- coding: utf-8 -*-
"""
Created on Mon Jan 15 14:06:54 2024

@author: peruzzetto
"""

import pytest
import numpy as np
import scipy
import os

import matplotlib.pyplot as plt

import tilupy.read as tiread


@pytest.fixture
def simple_temporal_results():
    d = np.ones((10, 15, 5))
    d[..., 0] = 0
    d[..., -1] = -1
    test_data = tiread.TemporalResults2D(
        "h",
        d,
        np.arange(5),
        x=np.arange(10),
        y=np.arange(15),
    )
    return test_data


@pytest.fixture
def gaussian_temporal_results():
    xmin, xmax = -1, 1
    ymin, ymax = -1, 1
    dx = dy = 0.05
    x = np.arange(xmin, xmax + dx / 2, dx)
    y = np.arange(ymin, ymax + dy / 2, dy)

    Dx = xmax - xmin
    Dy = ymax - ymin
    nx = len(x)
    ny = len(y)
    nt = 10

    xv, yv = np.meshgrid(x, y, indexing="xy")
    yv = np.flip(yv, axis=0)
    pos = np.dstack((xv, yv))
    h = np.zeros((ny, nx, nt))
    meanx = np.linspace(xmin + 0.2 * Dx, xmax - 0.4 * Dx, nt)
    meany = np.linspace(ymin + 0.3 * Dy, ymax - 0.3 * Dy, nt)
    stdx = np.linspace(0.02 * Dx, 0.08 * Dx, nt)
    stdy = [0.05 * Dy for i in range(nt)]
    cov_r = np.linspace(0, 0.5, nt)
    for i in range(nt):
        mean = [meanx[i], meany[i]]
        cov = cov_r[i] * np.sqrt(stdx[i] * stdy[i])
        cov_m = [[stdx[i], cov], [cov, stdy[i]]]
        rm = scipy.stats.multivariate_normal(mean, cov_m)
        h[:, :, i] = rm.pdf(pos)

    test_data = tiread.TemporalResults2D(
        "h",
        h,
        list(range(nt)),
        x=x,
        y=y,
    )

    return test_data


@pytest.mark.parametrize(
    "arg, expected",
    [
        ("int", ("h_int", 2, 2.5)),
        ("mean", ("h_mean", 2, 0.4)),
    ],
)
def test_get_temporal_stat(arg, expected, simple_temporal_results):
    res = simple_temporal_results.get_temporal_stat(arg)
    res_out = (res.name, res.d.ndim, res.d[0, 0])
    assert res_out == expected


@pytest.mark.parametrize(
    "args, expected",
    [
        (("int", "y"), ("h_int_y", 2, 0, 10)),
        (("int", "x"), ("h_int_x", 2, 0, 15)),
        (("mean", "x"), ("h_mean_x", 2, 0, 1)),
    ],
)
def test_get_spatial_stat(args, expected, simple_temporal_results):
    res = simple_temporal_results.get_spatial_stat(*args)
    res_out = (res.name, res.d.ndim, res.d[0, 0], res.d[0, 1])
    assert res_out == expected


@pytest.fixture(scope="function")
def plot_res():
    def _plot(res):
        res.plot()
        yield plt.show()
        plt.close("all")

    return _plot


@pytest.mark.parametrize(
    "args, expected",
    [
        (("int", "y"), None),
        (("int", "x"), None),
        (("int", "xy"), None),
    ],
)
def test_plot_spatial_stat(args, expected, gaussian_temporal_results, folder_plots):
    os.makedirs(folder_plots, exist_ok=True)
    file_out = "gaussian_{}_{}.png".format(args[0], args[1])
    file_out = os.path.join(folder_plots, file_out)
    if os.path.isfile(file_out):
        os.remove(file_out)
    res = gaussian_temporal_results.get_spatial_stat(*args)
    axe = res.plot()
    axe.figure.savefig(file_out)
    assert os.path.isfile(file_out)


def test_centermass(gaussian_temporal_results):
    simu = tiread.Results()
    simu._h = gaussian_temporal_results.d
    simu._zinit = gaussian_temporal_results.d[:, :, 0] * 0
    simu._tim = gaussian_temporal_results.t
    simu._x = gaussian_temporal_results.x
    simu._y = gaussian_temporal_results.y
    # print(simu._zinit.shape)
    # print(len(simu.x), len(simu.y))
    Dx = simu._x[-1] - simu._x[0]
    Dy = simu._y[-1] - simu._y[0]
    nt = len(simu.tim)
    res = simu.center_of_mass()
    meanx = np.linspace(simu.x[0] + 0.2 * Dx, simu.x[-1] - 0.4 * Dx, nt)
    meany = np.linspace(simu.y[0] + 0.3 * Dy, simu.x[-1] - 0.3 * Dy, nt)
    vol = np.sum(simu._h, axis=(0, 1))
    meanz = np.sum(simu._h**2 / 2, axis=(0, 1)) / vol
    maxdiffx = np.max(np.abs(res.d[0, :] - meanx))
    maxdiffy = np.max(np.abs(res.d[1, :] - meany))
    maxdiffz = np.max(np.abs(res.d[2, :] - meanz))
    # assert (res.d[0, 0] == meanx[0]) & (res.d[1, 0] == meany[0])
    assert (maxdiffx < 0.05) & (maxdiffy < 0.05) & (maxdiffz < 0.05)
    # assert simu._zinit.shape == (len(simu.x), len(simu.y))



# -*- coding: utf-8 -*-

import warnings
import numpy as np
import os
import shutil

from scipy.interpolate import RegularGridInterpolator

import tilupy.raster


def make_edges_matrices(nx: int, ny: int) -> list[np.ndarray, np.ndarray]:
    """Numbering edges for a regular rectangular grid.

    Considering a matrix M with whape (ny, nx),the convention is that
    M[0, 0] corresponds to the lower left corner of the matrix, M[0, -1]
    to the lower right corner, M[-1, 0] to the upper left and M[-1, -1] to
    the upper right. Edges are numbered cell by cell, counter clockwise :
    bottom, right, up, left. The first cell is M[0, 0], then cells are
    processed line by line.

    Parameters
    ----------
    nx : int
        Number of faces in the X direction
    ny : int
        Number of faces in the Y direction

    Returns
    -------
    h_edges : numpy.ndarray
        Matrix containing the global edge indices of all horizontal edges
        (bottom and top edges of the cells). Rows correspond to y-levels,
        columns to x-positions.
    v_edges : numpy.ndarray
        Matrix containing the global edge indices of all vertical edges
        (left and right edges of the cells). Rows correspond to y-levels,
        columns to x-positions.
    """
    # Numbers of horizontal edges
    h_edges = np.zeros((ny + 1, nx))
    # Numbers of vertical edges
    v_edges = np.zeros((ny, nx + 1))

    # Number of edges on first completed line
    n_edges_l1 = 4 + 3 * (nx - 1)
    # Number of edges on folowwing lines
    n_edges_l = 3 + 2 * (nx - 1)

    # Fill first line of h_edges
    h_edges[0, 0] = 1
    h_edges[0, 1:] = np.arange(nx - 1) * 3 + 5
    # Fill second line of h_edges
    h_edges[1, :] = h_edges[0, :] + 2

    try:
        # Fill first column of h_edges
        h_edges[2:, 0] = np.arange(ny - 1) * n_edges_l + n_edges_l1 + 2
        # Fill the rest of the table
        tmp = np.concatenate(([3], np.arange(nx - 2) * 2 + 5))
        h_edges[2:, 1:] = h_edges[2:, 0][:, np.newaxis] + tmp[np.newaxis, :]
    except IndexError:
        pass

    # Fill first  2 columns of v_edges
    v_edges[0, 0] = 4
    try:
        v_edges[1:, 0] = np.arange(ny - 1) * n_edges_l + n_edges_l1 + 3
    except IndexError:
        pass
    v_edges[:, 1] = v_edges[:, 0] - 2
    # Fill the rest of v_edges
    try:
        v_edges[0, 2:] = 2 + np.concatenate(([4], np.arange(nx - 2) * 3 + 7))
        tmp = np.concatenate(([3], np.arange(nx - 2) * 2 + 5))
        v_edges[1:, 2:] = v_edges[1:, 1][:, np.newaxis] + tmp[np.newaxis, :]
    except IndexError:
        pass

    h_edges = np.flip(h_edges, axis=0).astype(int)
    v_edges = np.flip(v_edges, axis=0).astype(int)
    return h_edges, v_edges


class ModellingDomain:
    """Mesh of simulation topography

    This class represents a structured 2D mesh built either from a raster
    or from user-specified dimensions and spacing.
    It stores the topographic data (z), the grid geometry (x, y, dx, dy, nx, ny),
    and the numbering of edges used for discretization of the domain.

    Parameters
    ----------
        raster : numpy.ndarray, str, or None, optional
            - If numpy.ndarray: used directly as the elevation grid (z).
            - If str: path to a raster file readable by :func:`tilupy.raster.read_raster`.
            - If None: grid is generated from :data:`nx`, :data:`ny`, :data:`dx`, :data:`dy`.
        xmin : float, optional
            Minimum X coordinate of the grid, used if :data:`raster` is None,
            by default 0.0.
        ymin : float, optional
            Minimum Y coordinate of the grid, used if :data:`raster` is None,
            by default 0.0.
        nx : int, optional
            Number of grid points in the X direction, used if :data:`raster` is None,
            by default None.
        ny : int, optional
            Number of grid points in the Y direction, used if :data:`raster` is None,
            by default None.
        dx : float, optional
            Grid spacing along X, by default 1.0.
        dy : float, optional
            Grid spacing along Y, by default 1.0.

    Attributes
    ----------
        _z : numpy.ndarray
            Elevation values of the mesh nodes ([ny, nx]).
        _nx : int
            Number of grid points along the X direction.
        _ny : int
            Number of grid points along the Y direction.
        _dx : float
            Grid resolution in the X direction.
        _dy : float
            Grid resolution in the Y direction.
        _x : numpy.ndarray
            Array of X coordinates.
        _y : numpy.ndarray
            Array of Y coordinates.
        _h_edges : numpy.ndarray
            Matrix of horizontal edge numbers ([ny, nx-1]).
        _v_edges : numpy.ndarray
            Matrix of vertical edge numbers ([ny-1, nx]).
    """

    def __init__(
        self,
        raster: np.ndarray = None,
        xmin: float = 0,
        ymin: float = 0,
        nx: int = None,
        ny: int = None,
        dx: float = 1,
        dy: float = 1,
    ):
        if raster is not None:
            if isinstance(raster, np.ndarray):
                self._z = raster
                self._nx = raster.shape[1]
                self._ny = raster.shape[0]
                self._dx = dx
                self._dy = dy
                self._x = np.arange(
                    xmin, xmin + (self._nx - 1) * self._dx + self._dx / 2, self._dx
                )
                self._y = np.arange(
                    ymin, ymin + (self._ny - 1) * self._dy + self._dy / 2, self._dy
                )

            if isinstance(raster, str):
                self._x, self._y, self._z = tilupy.raster.read_raster(raster)
                self._nx = len(self._x)
                self._ny = len(self._y)
                self._dx = self._x[1] - self._x[0]
                self._dy = self._y[1] - self._y[0]
        else:
            self._z = raster
            self._nx = nx
            self._ny = ny
            self._dx = dx
            self._dy = dy
            if (
                xmin is not None
                and ymin is not None
                and nx is not None
                and ny is not None
                and dx is not None
                and dy is not None
            ):
                self._x = np.arange(xmin, xmin + dx * nx + dx / 2, dx)
                self._y = np.arange(ymin, ymin + dy * ny + dy / 2, dy)

        self._h_edges = None
        self._v_edges = None

        if self._nx is not None and self._ny is not None:
            self.set_edges()

    def set_edges(self) -> None:
        """Compute and assign horizontal and vertical edge number.

        Calls :func:`tilupy.initsimus.make_edges_matrices` to build the edge numbering
        for the grid defined by nx and ny.
        Results are stored in attributes :attr:`_h_edges` and :attr:`_v_edges`.
        """
        self._h_edges, self._v_edges = make_edges_matrices(self._nx - 1, self._ny - 1)

    def get_edge(self, xcoord: float, ycoord: float, cardinal: str) -> int:
        """Get edge number for given coordinates and cardinal direction

        Parameters
        ----------
        xcoord : float
            X coordinate of the point.
        ycoord : float
            Y coordinate of the point.
        cardinal : str
            Cardinal direction of the edge ('N', 'S', 'E', 'W'):

                - 'N': top edge of the cell
                - 'S': bottom edge of the cell
                - 'E': right edge of the cell
                - 'W': left edge of the cell

        Returns
        -------
        int
            Edge number corresponding to the requested location and direction.

        Raises
        ------
        AssertionError
            If :meth:`set_edges` has not been called before (i.e. :attr:`_h_edges` and :attr:`_v_edges` are None).
        """
        assert (self._h_edges is not None) and (self._v_edges is not None), (
            "h_edges and v_edges must be computed to get edge number"
        )
        ix = np.argmin(np.abs(self._x[:-1] + self._dx / 2 - xcoord))
        iy = np.argmin(np.abs(self._y[:-1] + self._dx / 2 - ycoord))
        if cardinal == "S":
            return self._h_edges[-iy - 1, ix]
        elif cardinal == "N":
            return self._h_edges[-iy - 2, ix]
        elif cardinal == "W":
            return self._v_edges[-iy - 1, ix]
        elif cardinal == "E":
            return self._v_edges[-iy - 1, ix + 1]

    def get_edges(
        self,
        xcoords: np.ndarray,
        ycoords: np.ndarray,
        cardinal: str,
        from_extremities: bool = True,
    ) -> dict[str : list[int]]:
        """Get edges numbers for a set of coordinates and cardinals direction.

        If :data:`from_extremities` is True, xcoords and ycoords are treated as the extremities of a segment.

        Parameters
        ----------
        xcoords : np.ndarray or list
            X coordinates of points. If :data:`from_extremities` is True, indicated (xmin, xmax).
        ycoords : np.ndarray or list
            Y coordinates of points. If :data:`from_extremities` is True, indicated (ymin, ymax).
        cardinal : str
            Cardinal directions to extract, any combination of {'N', 'S', 'E', 'W'}.
            Example: "NS" will return both north and south edges.
        from_extremities : bool, optional
            If True, :data:`xcoords` and :data:`ycoords` are considered as the endpoints
            of a line, by default True

        Returns
        -------
        dict[str: list[int]]
            Dictionary where keys are the requested cardinals and values are
            sorted lists of unique edge numbers.
        """
        if from_extremities:
            d = np.sqrt((xcoords[1] - xcoords[0]) ** 2 + (ycoords[1] - ycoords[0]) ** 2)
            npts = int(np.ceil(d / min(self._dx, self._dy)))
            xcoords = np.linspace(xcoords[0], xcoords[1], npts + 1)
            ycoords = np.linspace(ycoords[0], ycoords[1], npts + 1)

        if self._h_edges is None or self._v_edges is None:
            self.set_edges

        res = dict()

        # cardinal is a string with any combination of cardinal directions
        for card in cardinal:
            edges_num = []
            for xcoord, ycoord in zip(xcoords, ycoords):
                edges_num.append(self.get_edge(xcoord, ycoord, card))
            res[card] = list(set(edges_num))  # Duplicate edges are removed
            res[card].sort()

        return res


class Simu:
    """Simulation configuration and input file generator.

    This class manages the setup of lave2D. It creates the necessary input files (topography, numeric
    parameters, rheology, boundary conditions, initial mass) that will be read by lave2D.

    Parameters
    ----------
        folder : str
            Directory where simulation files are written. Created if it does not exist.
        name : str
            Base name of the simulation (used as file prefix).

    Attributes
    ----------
        _folder : str
            Path to the directory where simulation input and output files are stored.
        _name : str
            Base name of the simulation, used as a prefix for generated files. Must have 8 characters.
        _x : numpy.ndarray
            Array of X coordinates of the topography grid.
        _y : numpy.ndarray
            Array of Y coordinates of the topography grid.
        _z : numpy.ndarray
            2D array of topographic elevations.
        _tmax : float
            Maximum simulation time.
        _dtsorties : float
            Time step for output results.
    """

    def __init__(self, folder: str, name: str):
        os.makedirs(folder, exist_ok=True)
        self._folder = folder
        self._name = name
        self._x = None
        self._y = None
        self._z = None

    def set_topography(
        self,
        z: np.ndarray | str,
        x: np.ndarray = None,
        y: np.ndarray = None,
        file_out: str = None,
    ) -> None:
        """Set simulation topography and write it as an ASCII grid.

        Parameters
        ----------
        z : ndarray or str
            Topography values as a 2D NumPy array, or path to a raster file.
        x : ndarray, optional
            X coordinates of the grid, ignored if :data:`z` is a file path.
        y : ndarray, optional
            Y coordinates of the grid, ignored if :data:`z` is a file path.
        file_out : str, optional
            Output filename (add ".asc"), by defaults "toposimu.asc".
            Filenames are truncated or padded to 8 characters.

        Returns
        -------
        None
        """
        if isinstance(z, str):
            self._x, self._y, self._z = tilupy.raster.read_raster(z)
        else:
            self._x, self._y, self._z = x, y, z

        if file_out is None:
            file_out = "toposimu.asc"
        else:
            fname = file_out.split(".")[0]
            if len(fname) > 8:
                warnings.warns("""Topography file name is too long,
                               only 8 first characters are retained""")
                fname = fname[:8]
            elif len(fname) < 8:
                warnings.warns("""Topography file name is too short and is adapted,
                               exactly 8 characters needed""")
                fname = fname.ljust(8, "0")
            file_out = fname + ".asc"

        tilupy.raster.write_ascii(
            self._x, self._y, self._z, os.path.join(self._folder, file_out)
        )

    def set_numeric_params(
        self,
        tmax: float,
        dtsorties: float,
        paray: float = 0.00099,
        dtinit: float = 0.01,
        cfl_const: int = 1,
        CFL: float = 0.2,
        alpha_cor_pente: float = 1.0,
    ) -> None:
        """Set numerical simulation parameters and write them to file.

        Parameters
        ----------
        tmax : float
            Maximum simulation time.
        dtsorties : float
            Time step for output results.
        paray : float, optional
            Hydraulic roughness parameter, by default 0.00099.
        dtinit : float, optional
            Initial time step, by default 0.01.
        cfl_const : int, optional
            If 1, CFL condition is constant, by default 1.
        CFL : float, optional
            Courant-Friedrichs-Lewy number, by default 0.2.
        alpha_cor_pente : float, optional
            Slope correction coefficient, by default 1.0.

        Returns
        -------
        None
        """
        self._tmax = tmax
        self._dtsorties = dtsorties

        with open(os.path.join(self._folder, "DONLEDD1.DON"), "w") as fid:
            fid.write("paray".ljust(34, " ") + "{:.12f}\n".format(paray).lstrip("0"))
            fid.write("tmax".ljust(34, " ") + "{:.12f}\n".format(tmax))
            fid.write("dtinit".ljust(34, " ") + "{:.12f}\n".format(dtinit))
            fid.write("cfl constante ?".ljust(34, " ") + "{:.0f}\n".format(cfl_const))
            fid.write("CFL".ljust(34, " ") + "{:.12f}\n".format(CFL))
            fid.write("Go,VaLe1,VaLe2".ljust(34, " ") + "{:.0f}\n".format(3))
            fid.write("secmbr0/1".ljust(34, " ") + "{:.0f}\n".format(1))
            fid.write("CL variable (si=0) ?".ljust(34, " ") + "{:d}\n".format(0))
            fid.write("dtsorties".ljust(34, " ") + "{:.12f}\n".format(dtsorties))
            fid.write(
                "alpha cor pentes".ljust(34, " ") + "{:.12f}".format(alpha_cor_pente)
            )

    def set_rheology(self, tau_rho: float, K_tau: float = 0.3) -> None:
        r"""Set Herschel-Bulkley rheology parameters

        Rheological parameters are tau/rho and K/tau. See :

            - Coussot, P., 1994. Steady, laminar, flow of concentrated mud suspensions in open channel. Journal of Hydraulic Research 32, 535-559. doi.org/10.1080/00221686.1994.9728354
                --> Eq 8 and text after eq 22 for the default value of K/tau
            - Rickenmann, D. et al., 2006. Comparison of 2D debris-flow simulation models with field events. Computational Geosciences 10, 241-264. doi.org/10.1007/s10596-005-9021-3
                --> Eq 9

        tau_rho : float
            Yield stress divided by density (:math:`\tau/\rho`).
        K_tau : float, optional
            Consistency index divided by yield stress (:math:`K/\tau`).
            By default 0.3, following Rickenmann (2006).

        Returns
        -------
        None
        """
        with open(os.path.join(self._folder, self._name + ".rhe"), "w") as fid:
            fid.write("{:.3f}\n".format(tau_rho))
            fid.write("{:.3f}".format(K_tau))

    def set_boundary_conditions(
        self,
        xcoords: list,
        ycoords: list,
        cardinal: str,
        discharges: list | float,
        times: list = None,
        thicknesses: list = None,
        tmax: float = 9999,
        discharge_from_volume: bool = False,
    ) -> None:
        r"""Define and write inflow hydrograph boundary conditions.

        Parameters
        ----------
        xcoords : list
            X coordinates of boundary location (two values for segment ends).
        ycoords : list
            Y coordinates of boundary location (two values for segment ends).
        cardinal : str
            Boundary orientation ('N', 'S', 'E', 'W').
        discharges : list or float
            If :data:`times` is None: interpreted as inflow volume (:math:`m^3`).
            Else: Interpreted as discharge values (:math:`m^3/s`) at each time step.
        times : list, optional
            Time vector associated with discharges. If None, a synthetic
            hydrograph is built from volume. By default None.
        thicknesses : list, optional
            Flow thicknesses for each time step. If None, put 1m everywhere. By default None.
        tmax : float, optional
            Maximum time for synthetic hydrograph generation, by default 9999.
        discharge_from_volume : bool, optional
            If True, discharges are computed from inflow volume, by default False. Not used.

        Returns
        -------
        None

        Raises
        ------
        AttributeError
            If no topography not been set.
        """
        try:
            modelling_domain = ModellingDomain(
                self._z,
                self._x[0],
                self._y[0],
                dx=self._x[1] - self._x[0],
                dy=self._y[1] - self._y[0],
            )
        except AttributeError:
            print("Simulation topography has not been set")
            raise

        edges = modelling_domain.get_edges(
            xcoords, ycoords, cardinal, from_extremities=True
        )
        edges = edges[cardinal]

        if times is None:
            # If no time vector is given, discharges is interpreted as a volume
            # from which an empirical peak discharge is computed
            peak_discharge = 0.0188 * discharges**0.79
            times = [0, 2 * discharges / peak_discharge, tmax]
            discharges = [peak_discharge, 0, 0]

        n_times = len(times)
        n_edges = len(edges)

        assert n_times == len(discharges), (
            "discharges and time vectors must be the same length"
        )

        if thicknesses is None:
            thicknesses = [1 for i in range(n_times)]

        if cardinal in ["W", "E"]:
            dd = self._y[1] - self._y[0]
        else:
            dd = self._x[1] - self._x[0]

        with open(os.path.join(self._folder, self._name + ".cli"), "w") as fid:
            fid.write("{:d}\n".format(n_times))  # Number of time steps in the hydrogram
            for i, time in enumerate(times):
                fid.write("{:.4f}\n".format(time))  # Write time
                fid.write("{:d}\n".format(n_edges))  # Write n_edges
                qn = discharges[i] / (n_edges * dd)
                qt = 0
                for i_edge, edge in enumerate(edges):
                    fid.write(
                        "\t{:d} {:.7E} {:.7E} {:.7E}\n".format(
                            edge, thicknesses[i], qn, qt
                        )
                    )

    def set_init_mass(self, mass_raster: str, h_min: float = 0.01) -> None:
        """Set initial debris-flow mass from a raster file.

        Reads an ASCII raster of initial thickness and interpolates it onto
        the simulation grid cell centers.

        Parameters
        ----------
        mass_raster : str
            Path to raster file containing initial mass thickness.
        h_min : float, optional
            Minimum non-zero thickness assigned to the grid. Default is 0.01.

        Returns
        -------
        None
        """
        x, y, m = tilupy.raster.read_ascii(mass_raster)
        # The input raster is given as the topography for the grid corners,
        # but results are then written on grid cell centers
        x2 = x[1:] - (x[1] - x[0]) / 2
        y2 = y[1:] - (y[1] - y[0]) / 2

        y2 = y2[::-1]

        fm = RegularGridInterpolator(
            (y, x), m, method="linear", bounds_error=False, fill_value=None
        )
        x_mesh, y_mesh = np.meshgrid(x2, y2)
        m2 = fm((y_mesh, x_mesh))
        m2[m2 < h_min] = h_min
        np.savetxt(
            os.path.join(self._folder, self._name + ".cin"),
            m2.flatten(),
            header="0.000000E+00",
            comments="",
            fmt="%.10E",
        )


def write_simu(
    raster_topo: str,
    raster_mass: str,
    tmax: float,
    dt_im: float,
    simulation_name: str,
    lave2D_exe_folder: str,
    rheology_type: str,
    rheology_params: dict,
    folder_out: str = None,
) -> None:
    """
    Prepares all input files required for a Lave2D simulation and saves them in a dedicated folder.

    Parameters
    ----------
    raster_topo : str
        Name of the ASCII topography file.
    raster_mass : str
        Name of the ASCII initial mass file.
    tmax : float
        Maximum simulation time.
    dt_im : float
        Output image interval (in time steps).
    simulation_name : str
        Simulation/project name.
    lave2D_exe_folder : str
        Path to the folder containing "Lave2_Arc.exe" and "vf2marc.exe".
    rheology_type : str
        Rheology to use for the simulation.
    rheology_params : dict
        Necessary parameters for the rheology. For this case:
            - tau_rho
            - K_tau
    folder_out : str, optional
        Output folder where simulation inputs will be stored.

    Returns
    -------
    None

    Raises
    ------
    ValueError
        If the rheology isn't Herschel_Bulkley.
    """
    if folder_out is None:
        folder_out = "."

    if rheology_type != "Herschel_Bulkley":
        raise ValueError("Rheology type must be 'Herschel_Bulkley'.")

    # output_file = os.path.join(folder_out, "lave2D")

    os.makedirs(folder_out, exist_ok=True)

    shutil.copy2(os.path.join(lave2D_exe_folder, "Lave2_Arc.exe"), folder_out)
    shutil.copy2(os.path.join(lave2D_exe_folder, "vf2marc.exe"), folder_out)

    simu_lave2D = Simu(folder_out, simulation_name)
    simu_lave2D.set_topography(raster_topo)
    simu_lave2D.set_init_mass(raster_mass)
    simu_lave2D.set_rheology(rheology_params["tau_rho"], rheology_params["K_tau"])
    simu_lave2D.set_numeric_params(tmax, dt_im)

    ## Not used in simulation, but the .cli file is needed
    simu_lave2D.set_boundary_conditions(
        [0, 0],  # X min and max coords for input flux
        [1, 2],  # Y min and max coord for input flux
        "W",  # Cardinal direction (Flow from East to West)
        [0, 0],  # Discharge hydrogram
        times=[0, tmax + 1],  # Corresponding times
        thicknesses=[0, 0],  # Thickness hydrogramm
    )


"""
if __name__ == "__main__":
    h_edges, v_edges = make_edges_matrices(1, 1)


if __name__ == "__main__":
    domain = ModellingDomain(xmin=0, ymin=0, nx=4, ny=3, dx=1, dy=1)
    domain.set_edges()
    res = domain.get_edges([1.5, 2.5], [0, 0], "S")
"""



# -*- coding: utf-8 -*-

from typing import Callable

import math as math
import numpy as np
import matplotlib
import matplotlib.pyplot as plt

import tilupy.read
import tilupy.notations as notations
from tilupy.analytic_sol import Coussot_shape


class Benchmark:
    """Benchmark of simulation results.

    This class groups together all the methods for processing and analyzing
    the results of various simulation models, allowing, among other things,
    the comparison of results between models.

    Attributes
    ----------
        _loaded_results : dict[tilupy.read.Results]
            Dictionnary of :class:`tilupy.read.Results` object for each model specified in
            :meth:`load_numerical_result`.
        _models_tim : dict[list]
            Dictionnary of recorded time steps for each model specified in
            :meth:`load_numerical_result`.
        _simu_z : numpy.ndarray
            Topographic elevation for the simulation loaded.
    """

    def __init__(self):
        self._loaded_results = {}
        self._models_tim = {}
        self._simu_z = None

    def load_numerical_result(
        self,
        model: str,
        **kwargs,
    ) -> None:
        """Load numerical simulation results using :func:`tilupy.read.get_results` for a given model and
        store the :class:`tilupy.read.Results` object in :attr:`_loaded_results`.

        Parameters
        ----------
        model : str
            Name of the model to load. Must be one of the :data:`tilupy.read.ALLOWED_MODELS`.
        **kwargs
            Keyword arguments passed to the result reader for the specific model.

        Raises
        ------
        ValueError
            If size of stored :attr:`_simu_z` is different with new loaded ones.
        UserWarning
            If stored :attr:`_simu_z` are different with new loaded ones.
        ValueError
            If the provided model is not in the allowed list of :data:`tilupy.read.ALLOWED_MODELS`.
        """
        if model in tilupy.read.ALLOWED_MODELS:
            if model not in self._loaded_results:
                self._loaded_results[model] = tilupy.read.get_results(model, **kwargs)
                self._models_tim[model] = self._loaded_results[model].tim

            if self._simu_z is None:
                self._simu_z = self._loaded_results[model].z

            if self._simu_z.shape != self._loaded_results[model].z.shape:
                raise ValueError("Simulations loaded doesn't have same size.")

            try:
                if not np.allclose(
                    self._simu_z, self._loaded_results[model].z, rtol=0.1, atol=0.1
                ):
                    mean_error = np.mean(self._simu_z - self._loaded_results[model].z)
                    self._simu_z = self._loaded_results[model].z
                    raise UserWarning(
                        f"Stored elevation values different from loaded ones; average error: {mean_error}"
                    )
            except UserWarning as w:
                print(f"[WARNING] {w}")

        else:
            raise ValueError(
                f" -> No correct model selected, choose between:\n       {tilupy.read.ALLOWED_MODELS}"
            )

    def compute_analytic_solution(
        self,
        output: str,
        solution: Callable,
        model: str,
        T: float | list[float],
        **kwargs,
    ) -> None:
        """
        Compute the analytic solution for the given time steps using the provided model.

        Parameters
        ----------
        output : str
            Wanted output for the analytical solution. Can be : "h" or "u".
        solution : Callable
            Callable object representing the analytic solution model (model from :class:`tilupy.analytic_sol.Depth_result`)
        model : str
            Wanted model to compute the analytical solution from. Must be loaded first with
            :meth:`load_numerical_result`.
        T : float | list[float]
            Time or list of times at which to compute the analytic solution.
        **kwargs
            Keyword arguments passed to the analytic solution for the specific model.

        Raises
        ------
        ValueError
            If no solution found.
        """
        solution = solution(**kwargs)

        if isinstance(T, float) or isinstance(T, int):
            T = [T]

        if output not in ["h", "u"]:
            raise ValueError(" -> Available output: 'h', 'u'.")

        if output == "h":
            solution.compute_h(self._loaded_results[model].x, T)

            if solution.h is not None:
                return tilupy.read.TemporalResults1D(
                    name=output,
                    d=solution.h[:].T,
                    t=T,
                    coords=self._loaded_results[model].x,
                    coords_name="x",
                )
            else:
                raise ValueError("No analytic solution for fluid height.")

        if output == "u":
            solution.compute_u(self._loaded_results[model].x, T)

            if solution.u is not None:
                return tilupy.read.TemporalResults1D(
                    name=output,
                    d=solution.u[:].T,
                    t=T,
                    coords=self._loaded_results[model].x,
                    coords_name="x",
                )
            else:
                raise ValueError("No analytic solution for fluid velocity.")

    def show_output(
        self,
        output: str,
        model: str,
        time_steps: float | list[float] = None,
        show_plot: bool = True,
        **plot_kwargs,
    ) -> matplotlib.axes._axes.Axes:
        """Plot output for a given model using :meth:`tilupy.read.Results.plot`.

        Parameters
        ----------
        output : str
            Wanted output, need to be in :data:`tilupy.read.DATA_NAMES`.
        model : str
           Wanted model to show the data field. Must be loaded with :meth:`load_numerical_result`
        time_steps : float | list[float]
            Value or list of time steps display output.
            If None displays the output for all recorded time steps in the model's result.
            By default None.
        show_plot : bool, optional
            If True, show the plot, by default True.

        Returns
        -------
        matplotlib.axes._axes.Axes
            The created plot.

        Raises
        ------
        ValueError
            If model is not loaded.
        """
        if model not in self._loaded_results.keys():
            raise ValueError(" -> First load model using load_numerical_result.")

        plot_kwargs = {} if plot_kwargs is None else plot_kwargs

        ax = self._loaded_results[model].plot(
            output=output, time_steps=time_steps, display_plot=show_plot, **plot_kwargs
        )
        if show_plot:
            plt.show()

        return ax

    def show_output_profile(
        self,
        output: str,
        model: str,
        extraction_method: str = "axis",
        extraction_params: dict = None,
        time_steps: float | list[float] = None,
        show_plot: bool = True,
        **plot_kwargs,
    ) -> matplotlib.axes._axes.Axes:
        """Plot profile from 2D temporal or static output for a given model using
        :meth:`tilupy.read.Results.plot_profile`.

        Parameters
        ----------
        output : str
            Wanted output, need to be in :data:`tilupy.read.TEMPORAL_DATA_2D` or :data:`tilupy.read.STATIC_DATA_2D`
        model : str
            Wanted model to show the output profile. Must be loaded with :meth:`load_numerical_result`
        extraction_method : str, optional
            Wanted profile extraction method. See :func:`tilupy.utils.get_profile`. By default "axis".
        extraction_params : dict, optional
            Profile extraction parameters. See :func:`tilupy.utils.get_profile`. Be default None.
        time_steps : float | list[float], optional
            Value or list of time steps required to extract and display profiles.
            If None displays the profiles for all recorded time steps in the model's result.
            Only available for models. By default None.
        show_plot : bool, optional
            If True, show the plot, by default True.

        Returns
        -------
        matplotlib.axes._axes.Axes
            The created plot.

        Raises
        ------
        ValueError
            If model is not loaded.
        """
        if model not in self._loaded_results.keys():
            raise ValueError(" -> First load model using load_numerical_result")

        plot_kwargs = {} if plot_kwargs is None else plot_kwargs

        ax = self._loaded_results[model].plot_profile(
            output=output,
            extraction_method=extraction_method,
            extraction_params=extraction_params,
            time_steps=time_steps,
            display_plot=show_plot,
            **plot_kwargs,
        )
        if show_plot:
            plt.show()

        return ax

    def show_comparison_temporal1D(
        self,
        output: str,
        models: list[str],
        profile_extraction_args: dict = None,
        analytic_solution: dict = None,
        time_steps: float | list[float] = None,
        axes: matplotlib.axes._axes.Axes = None,
        rows_nb: int = None,
        cols_nb: int = None,
        figsize: tuple[float] = None,
        colors: list = None,
        linestyles: list[str] = None,
        plot_kwargs: dict = None,
        as_kwargs: dict = None,
        show_plot: bool = True,
    ) -> matplotlib.axes._axes.Axes:
        """Plot multiple temporal 1D data in the same graph.

        Parameters
        ----------
        output : str
            Wanted output, need to be in :data:`tilupy.read.TEMPORAL_DATA_1D`. If using a :data:`tilupy.read.TEMPORAL_DATA_2D`,
            must use :data:`profile_extraction_args`.
        models : list[str]
            Wanted models to show the temporal 1D data. Must be loaded with :meth:`load_numerical_result`.
        profile_extraction_args : dict, optional
            Arguments for profile extraction. See :func:`tilupy.utils.get_profile`. By default None.
        analytic_solution : dict, optional
            Arguments for plotting an analytic solution. See :meth:`compute_analytic_solution`.
            By default None.
        time_steps : float | list[float], optional
            Value or list of time steps required to extract and display profiles.
            If None displays the profiles for all recorded time steps in the model's result.
            Only available for models. By default None.
        axes : matplotlib.axes._axes.Axes, optional
            Existing matplotlib axe, by default None.
        rows_nb : int, optional
            Number of rows when plotting multiple time steps. If None automatically computed.
            By default None.
        cols_nb : int, optional
            Number of colums when plotting multiple time steps. If None use 3 columns.
            By default None.
        figsize : tuple[float], optional
            Size of the plotted figure (Height, Width), by default None.
        colors : list, optional
            List of colors for each model plotted. If None use "black". By default None.
        linestyles : list[str], optional
            List of linestyles for each model plotted. If None use regular straight line. By default None.
        plot_kwargs : dict, optional
            Additional argument for model's line, by default None.
        as_kwargs : dict, optional
            Additional argument for analytical solution's line, by default None.
        show_plot : bool, optional
            If True, show the plot, by default True.

        Returns
        -------
        matplotlib.axes._axes.Axes
            The created plot.

        Raises
        ------
        ValueError
            If models is not loaded.
        """
        for model in models:
            if model not in self._loaded_results.keys():
                raise ValueError(" -> First load model using load_numerical_result.")

        plot_kwargs = {} if plot_kwargs is None else plot_kwargs
        as_kwargs = {} if as_kwargs is None else as_kwargs
        profile_extraction_args = (
            {} if profile_extraction_args is None else profile_extraction_args
        )
        if "extraction_method" not in profile_extraction_args:
            profile_extraction_args["extraction_method"] = "axis"

        profile_models = {}
        for model in models:
            if output in tilupy.read.TEMPORAL_DATA_2D:
                prof, data = self._loaded_results[model].get_profile(
                    output=output, **profile_extraction_args
                )
                profile_models[model] = [prof, data]
            elif output in tilupy.read.TEMPORAL_DATA_1D:
                prof = self._loaded_results[model].get_output(output)
                profile_models[model] = [prof, None]

        if time_steps is None:
            time_steps = self._models_tim[list(self._loaded_results.keys())[0]]

        if analytic_solution is not None:
            as_profile = self.compute_analytic_solution(
                output=output,
                model=list(self._loaded_results.keys())[0],
                T=time_steps,
                **analytic_solution,
            )

        if isinstance(time_steps, float) or isinstance(time_steps, int):
            time_steps = np.array([time_steps])
        elif isinstance(time_steps, list):
            time_steps = np.array(time_steps)

        if isinstance(profile_models[model][0], tilupy.read.TemporalResults):
            for model in models:
                profile_models[model][0] = profile_models[model][
                    0
                ].extract_from_time_step(time_steps)

        if axes is None:
            if cols_nb is None:
                cols_nb = len(time_steps) if len(time_steps) < 3 else 3

            if rows_nb is None:
                rows_nb = len(time_steps) // cols_nb
                if len(time_steps) % cols_nb != 0:
                    rows_nb += 1

            fig, axes = plt.subplots(
                nrows=rows_nb,
                ncols=cols_nb,
                figsize=figsize,
                layout="constrained",
                sharex=True,
                sharey=True,
            )
            if isinstance(axes, plt.Axes):
                axes = np.array([axes])
            else:
                axes = axes.flatten()

        for T in range(len(time_steps)):
            # Plot models
            for i in range(len(models)):
                if colors is not None and len(colors) > i:
                    color = colors[i]
                else:
                    color = "black"
                if linestyles is not None and len(linestyles) > i:
                    linestyle = linestyles[i]
                else:
                    linestyle = None

                if "alpha" not in plot_kwargs:
                    plot_kwargs["alpha"] = 0.8
                if "linewidth" not in plot_kwargs:
                    plot_kwargs["linewidth"] = 1.5

                axes[T].plot(
                    profile_models[models[i]][0].coords,
                    profile_models[models[i]][0].d[:, T],
                    color=color,
                    linestyle=linestyle,
                    label=models[i],
                    **plot_kwargs,
                )

            # Plot analytic solution
            if analytic_solution is not None:
                if "color" not in as_kwargs:
                    as_kwargs["color"] = "red"
                if "alpha" not in as_kwargs:
                    as_kwargs["alpha"] = 0.9
                if "linewidth" not in as_kwargs:
                    as_kwargs["linewidth"] = 1

                axes[T].plot(
                    as_profile.coords,
                    as_profile.d[:, T],
                    label=f"{str(analytic_solution['solution']).split('.')[-1][:-2]}",
                    **as_kwargs,
                )

            # Formatting fig
            axes[T].set_xlim(
                left=min(profile_models[models[0]][0].coords),
                right=max(profile_models[models[0]][0].coords),
            )
            axes[T].grid(True, alpha=0.3)

            if len(time_steps) == 1:
                if (
                    profile_extraction_args["extraction_method"] == "axis"
                    and output in tilupy.read.TEMPORAL_DATA_2D
                ):
                    inv_axis = ""
                    if profile_models[models[0]][0].coords_name == "x":
                        inv_axis = "Y"
                    else:
                        inv_axis = "X"
                    axes[T].set_title(
                        f"{inv_axis}={profile_models[models[0]][1]}m  |  t={time_steps[T]}s"
                    )
                else:
                    axes[T].set_title(f"t={time_steps[T]}s")
            else:
                axes[T].set_title(f"t={time_steps[T]}s", loc="left")
                if (
                    profile_extraction_args["extraction_method"] == "axis"
                    and output in tilupy.read.TEMPORAL_DATA_2D
                ):
                    inv_axis = ""
                    if profile_models[models[0]][0].coords_name == "x":
                        inv_axis = "Y"
                    else:
                        inv_axis = "X"
                    axes[T].set_title(
                        f"{inv_axis}={profile_models[models[0]][1]}m", loc="right"
                    )

            axes[T].set_xlabel(
                notations.get_label(profile_models[models[0]][0].coords_name)
            )
            axes[T].set_ylabel(notations.get_label(output))
            axes[T].legend(loc="upper right")

        for i in range(len(time_steps), len(axes)):
            fig.delaxes(axes[i])

        if show_plot:
            plt.show()

        return axes

    def get_avrg_result(
        self, output: str
    ) -> tilupy.read.TemporalResults | tilupy.read.StaticResults:
        """Get average result computed with all loaded model (:meth:`load_numerical_result`).

        Parameters
        ----------
        output : str
            Wanted average output.

        Returns
        -------
        tilupy.read.TemporalResults | tilupy.read.StaticResults
            Average result output.

        Raises
        ------
        ValueError
            If StaticResults0D output wanted.
        """
        output_list = []
        for model in self._loaded_results:
            output_list.append(self._loaded_results[model].get_output(output))

        if isinstance(output_list[0], tilupy.read.TemporalResults2D):
            data_list = [output_list[i].d for i in range(len(output_list))]
            time_list = [output_list[i].t for i in range(len(output_list))]
            x_list = [output_list[i].x for i in range(len(output_list))]
            y_list = [output_list[i].y for i in range(len(output_list))]
            z_list = [output_list[i].z for i in range(len(output_list))]

            mean_data = np.mean(np.stack(data_list), axis=0)
            mean_time = np.mean(np.stack(time_list), axis=0)
            mean_x = np.mean(np.stack(x_list), axis=0)
            mean_y = np.mean(np.stack(y_list), axis=0)
            mean_z = np.mean(np.stack(z_list), axis=0)

            mean_result = tilupy.read.TemporalResults2D(
                name=output_list[0].name,
                d=mean_data,
                t=mean_time,
                x=mean_x,
                y=mean_y,
                z=mean_z,
            )

            return mean_result

        elif isinstance(output_list[0], tilupy.read.TemporalResults1D):
            data_list = [output_list[i].d for i in range(len(output_list))]
            time_list = [output_list[i].t for i in range(len(output_list))]
            coords_list = [output_list[i].coords for i in range(len(output_list))]

            mean_data = np.mean(np.stack(data_list), axis=0)
            mean_time = np.mean(np.stack(time_list), axis=0)
            mean_coords = np.mean(np.stack(coords_list), axis=0)

            mean_result = tilupy.read.TemporalResults1D(
                name=output_list[0].name,
                d=mean_data,
                t=mean_time,
                coords=mean_coords,
                coords_name=output_list[0].coords_name,
            )

            return mean_result

        elif isinstance(output_list[0], tilupy.read.TemporalResults0D):
            data_list = [output_list[i].d for i in range(len(output_list))]
            time_list = [output_list[i].t for i in range(len(output_list))]

            mean_data = np.mean(np.stack(data_list), axis=0)
            mean_time = np.mean(np.stack(time_list), axis=0)

            mean_result = tilupy.read.TemporalResults0D(
                name=output_list[0].name,
                d=mean_data,
                t=mean_time,
                scalar_names=output_list[0].scalar_names,
            )

            return mean_result

        elif isinstance(output_list[0], tilupy.read.StaticResults2D):
            data_list = [output_list[i].d for i in range(len(output_list))]
            x_list = [output_list[i].x for i in range(len(output_list))]
            y_list = [output_list[i].y for i in range(len(output_list))]
            z_list = [output_list[i].z for i in range(len(output_list))]

            mean_data = np.mean(np.stack(data_list), axis=0)
            mean_x = np.mean(np.stack(x_list), axis=0)
            mean_y = np.mean(np.stack(y_list), axis=0)
            mean_z = np.mean(np.stack(z_list), axis=0)

            mean_result = tilupy.read.StaticResults2D(
                name=output_list[0].name, d=mean_data, x=mean_x, y=mean_y, z=mean_z
            )

            return mean_result

        elif isinstance(output_list[0], tilupy.read.StaticResults1D):
            data_list = [output_list[i].d for i in range(len(output_list))]
            coords_list = [output_list[i].coords for i in range(len(output_list))]

            mean_data = np.mean(np.stack(data_list), axis=0)
            mean_coords = np.mean(np.stack(coords_list), axis=0)

            mean_result = tilupy.read.StaticResults1D(
                name=output_list[0].name,
                d=mean_data,
                coords=mean_coords,
                coords_name=output_list[0].coords_name,
            )

            return mean_result

        else:
            raise ValueError("Not available for StaticResults0D.")

    def compute_area(self, flow_threshold: float = None) -> tuple[dict, dict]:
        """Compute area at each recorded time steps, computed with 'h', for each model loaded
        using :meth:`load_numerical_result`.

        Parameters
        ----------
        flow_threshold : float, optional
            Flow threshold to extract flow area, if None use 1% of initial maximal
            flow height. By default None.

        Returns
        -------
        tuple[dict, dict]
            area_surf: dict
                2D area for each model: area_surf[model] = TemporalResults2D.
            area_num: dict
                Area value for each model: area_num[model] = TemporalResults0D.
        """
        output_list = []
        for model in self._loaded_results:
            output_list.append(self._loaded_results[model].get_output("h"))

        height_list = [output_list[i].d for i in range(len(output_list))]

        if flow_threshold is None:
            flow_threshold = np.max(height_list[0]) * 0.01

        for h in height_list:
            h[h < flow_threshold] = 0
            h[h >= flow_threshold] = 1

        area_num = {}
        area_surf = {}
        model_name = list(self._loaded_results.keys())
        for i in range(len(height_list)):
            surface_list = []
            dx = self._loaded_results[model].dx
            dy = self._loaded_results[model].dy
            cell_surface = dx * dy

            for t in range(height_list[i].shape[2]):
                nb_cell = np.sum(height_list[i][:, :, t] == 1)
                surface_list.append(nb_cell * cell_surface)

            area_num[model_name[i]] = tilupy.read.TemporalResults0D(
                name="s",
                d=np.array(surface_list),
                t=output_list[i].t,
                scalar_names="Surface",
            )

            area_surf[model_name[i]] = tilupy.read.TemporalResults2D(
                name="s",
                d=height_list[i],
                t=output_list[i].t,
                x=output_list[i].x,
                y=output_list[i].y,
                z=output_list[i].z,
            )

        return area_surf, area_num

    def compute_impacted_area(
        self,
        flow_threshold: float = None,
    ) -> tuple[dict, dict]:
        """Compute impacted area, computed with 'h_max', for each model loaded using
        :meth:`load_numerical_result`.

        Parameters
        ----------
        flow_threshold : float, optional
            Flow threshold to extract flow area, if None use 1% of initial maximal
            flow height. By default None.

        Returns
        -------
        tuple[dict, dict]
            area_surf: dict
                2D area for each model: area_surf[model] = StaticResults2D.
            area_num: dict
                Area value for each model: area_num[model] = area_val.
        """
        output_list = []
        for model in self._loaded_results:
            output_list.append(self._loaded_results[model].get_output("h_max"))

        height_list = [output_list[i].d for i in range(len(output_list))]

        if flow_threshold is None:
            flow_threshold = np.max(height_list[0]) * 0.01

        for h in height_list:
            h[h < flow_threshold] = 0
            h[h >= flow_threshold] = 1

        area_num = {}
        area_surf = {}
        model_name = list(self._loaded_results.keys())
        for i in range(len(height_list)):
            dx = self._loaded_results[model].dx
            dy = self._loaded_results[model].dy
            cell_surface = dx * dy

            nb_cell = np.sum(height_list[i][:, :] == 1)
            surface = nb_cell * cell_surface

            area_num[model_name[i]] = surface

            area_surf[model_name[i]] = tilupy.read.StaticResults2D(
                name="s",
                d=height_list[i],
                x=output_list[i].x,
                y=output_list[i].y,
                z=output_list[i].z,
            )

        return area_surf, area_num

    def compute_impacted_area_rms_from_avrg(
        self, flow_threshold: float = None
    ) -> tuple[dict, np.ndarray]:
        """Compute impacted area, computed with 'h_max', RMS with average result for each model loaded
        using :meth:`load_numerical_result`.

        Parameters
        ----------
        flow_threshold : float, optional
            Flow threshold to extract flow area, if None use 1% of initial maximal
            flow height. By default None.

        Returns
        -------
        tuple[dict, np.ndarray]
            area_rms: dict
                RMS value for each model: area_rms[model] = rms_value.
            avrg_area: numpy.ndarray
                Average area surface.
        """
        area_surf, area_num = self.compute_impacted_area(flow_threshold=flow_threshold)

        mean_area = []
        for model in area_surf:
            mean_area.append(np.nan_to_num(area_surf[model].d))

        avrg_area = np.mean(np.array(mean_area), axis=0)

        area_rms = {}
        for model in area_num:
            rms = np.sqrt(np.sum((area_surf[model].d - avrg_area) ** 2)) / np.sqrt(
                np.sum((avrg_area) ** 2)
            )
            area_rms[model] = rms

        return area_rms, avrg_area

    def compute_rms_from_avrg(
        self,
        output: str,
    ) -> dict:
        """Compute RMS with average result for each model loaded using :meth:`load_numerical_result`.

        Parameters
        ----------
        output : str
            Wanted output to compute the RMS.

        Returns
        -------
        dict
            RMS value for each model: output_rms[model] = rms_value.
        """
        avrg_result = self.get_avrg_result(output)

        extracted_data = {}
        for model in self._loaded_results:
            data = self._loaded_results[model].get_output(output)
            extracted_data[model] = data.d
        avrg_height = avrg_result.d

        output_rms = {}
        for model in self._loaded_results:
            rms = np.sqrt(np.sum((extracted_data[model] - avrg_height) ** 2)) / np.sqrt(
                np.sum((avrg_height) ** 2)
            )
            output_rms[model] = rms

        return output_rms

    def compute_dist_centermass(
        self, flow_threshold: float = None
    ) -> tuple[dict, dict]:
        """Compute the distance between the initial position of the center of mass and the final furthest point
        for each model loaded using :meth:`load_numerical_result`.

        Parameters
        ----------
        flow_threshold : float, optional
            Flow threshold to extract flow area, if None use 1% of initial maximal
            flow height. By default None.

        Returns
        -------
        tuple[dict, dict]
            init_coord_centermass : dict
                Coordinates of the initial center of mass for each model:
                init_coord_centermass[model] = [X, Y, Z].
            model_max_dist : dict
                Maximal distance for each model : model_max_dist[model] = dist.
        """
        init_coord_centermass = {}
        for model in self._loaded_results:
            centermass = self._loaded_results[model].get_output("centermass")
            init_coord_centermass[model] = [
                centermass.d[0, 0],
                centermass.d[1, 0],
                centermass.d[2, 0],
            ]

        area_surf, _ = self.compute_impacted_area(flow_threshold=flow_threshold)

        model_max_dist = {}
        for model in self._loaded_results:
            model_max_dist[model] = [None, None, 0.0]

        for model in self._loaded_results:
            for i in range(len(area_surf[model].x)):
                for j in range(len(area_surf[model].y)):
                    if area_surf[model].d[j, i] == 1:
                        distance = np.sqrt(
                            (area_surf[model].x[i] - init_coord_centermass[model][0])
                            ** 2
                            + (area_surf[model].y[j] - init_coord_centermass[model][1])
                            ** 2
                            + (
                                area_surf[model].z[j, i]
                                - init_coord_centermass[model][2]
                            )
                            ** 2
                        )
                        if distance > model_max_dist[model][2]:
                            model_max_dist[model][0] = area_surf[model].x[i]
                            model_max_dist[model][1] = area_surf[model].y[j]
                            model_max_dist[model][2] = distance

        return init_coord_centermass, model_max_dist

    def compute_average_velocity(
        self,
        distance: float = None,
        look_up_direction: str = "right",
        flow_threshold: float = None,
        **extration_profile_params,
    ) -> tuple[dict, dict, float, dict]:
        """Compute average velocity for each model loaded using :meth:`load_numerical_result`.

        Parameters
        ----------
        distance : float, optional
            Distance used to calculate average speed, if None use :data:`maximal_distance/2`.
            By default None.
        look_up_direction : str, optional
            Direction to look for the flow front, must be "right" or "left",
            by default "right".
        flow_threshold : float, optional
            Flow threshold when extracting front position from profile, if None use
            1% of initial maximal flow height. By default None.

        Returns
        -------
        tuple[dict, dict, float, dict]
            model_avrg_vel: dict
                Average velocity value for each model: model_avrg_vel[model] = avrg_velocity.
            model_time: dict
                Time for each model to reach the distance: model_time[model] = time.
            distance: float
                Distance used to compute the average velocity.
            model_pos: dict
                Front position of the profile of each model: model_pos[model] = [init_front_pos, final_front_pos,
                maximal_distance].
        """

        def get_front_index(profile):
            idx = np.where(profile <= flow_threshold)[0]
            return (idx[0] - 1) if len(idx) else len(profile) - 1

        def get_back_index(profile):
            idx = np.where(profile <= flow_threshold)[0]
            return (idx[-1] + 1) if len(idx) else 0

        extration_profile_params = (
            {} if extration_profile_params is None else extration_profile_params
        )

        if look_up_direction not in ["right", "left"]:
            raise ValueError("Invalid look-up direction: 'right' or 'left'")

        # Extract profile
        model_profile = {}
        for model in self._loaded_results:
            profile, _ = self._loaded_results[model].get_profile(
                output="h", **extration_profile_params
            )
            model_profile[model] = profile

            if flow_threshold is None:
                flow_threshold = np.max(profile.d[:, 0]) * 0.01

        # Find initial and final position
        model_pos = {}
        for model in self._loaded_results:
            max_index_init = np.argmax(model_profile[model].d[:, 0])
            max_index_final = np.argmax(model_profile[model].d[:, -1])

            idx_r_init = (
                get_front_index(model_profile[model].d[max_index_init:, 0])
                + max_index_init
            )
            idx_l_init = get_back_index(model_profile[model].d[:max_index_init, 0])
            idx_r_final = (
                get_front_index(model_profile[model].d[max_index_final:, -1])
                + max_index_final
            )
            idx_l_final = get_back_index(model_profile[model].d[:max_index_final, -1])

            if look_up_direction == "right":
                init_pos = model_profile[model].coords[idx_r_init]
                final_pos = model_profile[model].coords[idx_r_final]
            else:
                init_pos = model_profile[model].coords[idx_l_init]
                final_pos = model_profile[model].coords[idx_l_final]

            model_pos[model] = (init_pos, final_pos, abs(final_pos - init_pos))

        # Find minimal distance
        min_distance = 1e10
        for model in model_pos:
            min_distance = (
                model_pos[model][2]
                if model_pos[model][2] < min_distance
                else min_distance
            )

        if distance is not None and distance > min_distance:
            raise ValueError(
                f"The requested distance is greater than the minimum distance: {min_distance}"
            )

        if distance is None:
            distance = min_distance / 2

        # Find the time for each model to exceed the distance
        model_time = {}
        for model in model_profile:
            for t in range(1, len(model_profile[model].t)):
                max_index = np.argmax(model_profile[model].d[:, t])

                idx_r = (
                    get_front_index(model_profile[model].d[max_index:, t]) + max_index
                )
                idx_l = get_back_index(model_profile[model].d[:max_index, t])

                if look_up_direction == "right":
                    pos = model_profile[model].coords[idx_r]
                else:
                    pos = model_profile[model].coords[idx_l]

                if abs(pos - model_pos[model][0]) >= distance:
                    model_time[model] = model_profile[model].t[t]
                    break

        # Compute the avrg velocity for each model
        model_avrg_vel = {}
        for model in model_time:
            model_avrg_vel[model] = distance / model_time[model]

        return model_avrg_vel, model_time, distance, model_pos

    def compute_rms_from_coussot(
        self,
        coussot_params: dict,
        look_up_direction: str = "right",
        flow_threshold: float = None,
        **extration_profile_params,
    ) -> tuple[dict, dict, dict]:
        """Compute RMS with Coussot's front shape for each model loaded using :meth:`load_numerical_result`.

        Parameters
        ----------
        coussot_params : dict
            Arguments for generating Coussot's solution. See :class:`tilupy.analytic_sol.Coussot_shape`.
        look_up_direction : str, optional
            Direction to look for the flow front, must be "right" or "left",
            by default "right".
        flow_threshold : float, optional
            Flow threshold when extracting front position from profile, if None use
            1% of initial maximal flow height. By default None.

        Returns
        -------
        tuple[dict, dict, dict]
            output_rms: dict
                RMS value for each model: output_rms[model] = rms_value.
            model_front_pos: dict
                Position of Coussot's profile for each model: model_front_pos[model] = pos_value.
            model_coussot: dict
                Coussot's profile for each model: model_coussot[model] = StaticResults1D.
        """

        def get_front_index(profile):
            idx = np.where(profile <= flow_threshold)[0]
            return (idx[0] - 1) if len(idx) else len(profile) - 1

        def get_back_index(profile):
            idx = np.where(profile <= flow_threshold)[0]
            return (idx[-1] + 1) if len(idx) else 0

        output_rms = {}
        model_front_pos = {}
        model_coussot = {}
        for model in self._loaded_results:
            # Extract profile
            prof, _ = self._loaded_results[model].get_profile(
                output="h", **extration_profile_params
            )
            if flow_threshold is None:
                flow_threshold = np.max(prof.d[:, 0]) * 0.01
            max_index = np.argmax(prof.d[:, -1])
            # Create Coussot profile
            front_shape = Coussot_shape(**coussot_params, h_final=np.max(prof.d[:, -1]))
            front_shape.compute_rheological_test_front_morpho()

            if look_up_direction == "right":
                front_shape.change_orientation_flow()
                idx = get_front_index(prof.d[max_index:, -1]) + max_index
                front_pos = prof.coords[idx]
            else:
                idx = get_back_index(prof.d[:max_index, -1])
                front_pos = prof.coords[idx]
            front_shape.translate_front(front_pos)
            front_shape.interpolate_on_d()

            coussot_pos = front_shape.d.copy()

            # Find best position for Coussot's profile
            max_dx = (prof.coords[1] - prof.coords[0]) * 5
            max_index = np.argmax(prof.d[:, -1])

            dx_range = np.linspace(-max_dx, max_dx, 5)
            best_dx = 0.0
            min_rms = 1e30
            # x_index = [None, None]

            for dx in dx_range:
                temp_pos = coussot_pos + dx

                # Extract front value of profile
                if look_up_direction == "right":
                    x_index_max = (
                        np.argmin(np.abs(prof.coords[max_index:] - np.max(temp_pos)))
                        + max_index
                    )
                    x_index_min = (
                        np.argmin(np.abs(prof.coords[max_index:] - np.min(temp_pos)))
                        + max_index
                    )
                else:
                    x_index_max = np.argmin(
                        np.abs(prof.coords[:max_index] - np.max(temp_pos))
                    )
                    x_index_min = np.argmin(
                        np.abs(prof.coords[:max_index] - np.min(temp_pos))
                    )

                reduc_idx = np.linspace(
                    0,
                    len(temp_pos) - 1,
                    len(prof.coords[x_index_min:x_index_max]),
                    dtype=int,
                )

                rms = np.sqrt(
                    np.mean(
                        (
                            (prof.d[x_index_min:x_index_max, -1])
                            - (front_shape.h[reduc_idx])
                        )
                        ** 2
                    )
                )

                rms /= np.sqrt(np.mean((front_shape.h[reduc_idx]) ** 2))

                if rms <= min_rms:
                    min_rms = rms
                    best_dx = dx
                    # x_index = [x_index_min, x_index_max]

            output_rms[model] = min_rms
            front_shape.translate_front(best_dx)
            model_front_pos[model] = front_pos + best_dx

            model_coussot[model] = tilupy.read.StaticResults1D(
                name="h", d=front_shape.h, coords=front_shape.d, coords_name="x"
            )

            # prof.plot()
            # plt.plot(front_shape.d[reduc_idx], front_shape.h[reduc_idx], color='red')
            # plt.show()

            # plt.plot([0, np.max(front_shape.h[reduc_idx])], [0, np.max(front_shape.h[reduc_idx])], color='red')
            # plt.plot(prof.d[x_index[0]:x_index[1], -1], front_shape.h[reduc_idx])
            # plt.xlabel("Solution numérique")
            # plt.ylabel("Solution de Coussot")
            # plt.show()

        return output_rms, model_front_pos, model_coussot

    def generate_simulation_comparison_csv(
        self,
        save: bool = False,
        folder_out: str = None,
        file_name: str = None,
        fmt: str = "csv",
        available_profile: bool = False,
        extration_profile_params: dict = None,
        flow_threshold: float = 1e-1,
        profile_direction: str = "right",
        avrg_velocity_distance: float = None,
        coussot_criteria: dict = None,
    ) -> None:
        """Generate a csv file summarizing comparison criteria between flow models.

        Generates a file grouping comparison criteria between numerical flow models:

            - Criteria integrated throughout the simulation:

                - Flow Area: flow area value and difference with mean result (:meth:`compute_impacted_area`).
                - Impacted Zone: RMS of impacted area versus mean result (:meth:`compute_impacted_area_rms_from_avrg`).
                - Maximal Height: RMS of flow maximal height versus mean result (:meth:`compute_rms_from_avrg`).
                - Maximal Momentum: RMS of flow maximal mementum versus mean result (:meth:`compute_rms_from_avrg`).
                - Average Velocity (if profile): time for flow to reach a given distance
                and average velocity calculated from these values (:meth:`compute_average_velocity`).

            - Criteria for the final time step of the simulation:

                - Final Height: RMS of flow final height versus mean result (:meth:`compute_rms_from_avrg`).
                - Maximal Extension: distance from the initial center of mass and the final furthest point of the flow (:meth:`compute_dist_centermass`).
                - Flow Front Position (if profile): maximum distance traveled by the flow and comparison with the average result (:meth:`compute_average_velocity`).
                - Front Shape (optional): RMS of the front shape versus Coussot's theorical front shape (:meth:`compute_rms_from_coussot`).

            - Numerical criteria:

                - Volume: value of the volume at final time steps (compared to initial volume) and RMS versus initial volume value.

        Parameters
        ----------
        save : bool, optional
            If True, save the resulting tab at :data:`folder_out`. By default False.
        folder_out : str, optional
            Path to the folder where the file is saved, if None generate "xlsx_results"
            folder in code folder. By default None
        file_name : str, optional
            Name of the folder, if None use "results_[models]". By default None.
        fmt : str, optional
            Saving format of the table. Can be "csv" or "xlsx". By default "csv".
        available_profile: bool, optional
            If True, calculate criteria requiring a profile: avrg velocity and final front
            position. By default False.
        flow_threshold : float, optional
            Flow threshold when extracting front position from profile, by default 1e-1.
        profile_direction : str, optional
            Direction to look for the flow front, must be "right" or "left",
            by default "right".
        avrg_velocity_distance : float, optional
            Distance used to calculate average speed, if None use :data:`maximal_distance/2`.
            By default None.
        coussot_criteria : dict, optional
            If None, ignore Coussot criteria. Otherwise, list of arguments to generate a
            Coussot profile, by default None

        Raises
        ------
        ValueError
            If the file already exists.
        """
        import pandas as pd
        import openpyxl
        import os

        if save:
            if folder_out is None:
                folder_out = os.path.join(os.path.dirname(__file__), "xlsx_results")

            os.makedirs(folder_out, exist_ok=True)

            if file_name is None:
                file_name = "results"
                for model in self._loaded_results:
                    file_name += "_" + model

            if not file_name.endswith("." + fmt):
                file_name = file_name + "." + fmt
            saving_path = os.path.join(folder_out, file_name)

            if os.path.exists(saving_path):
                raise ValueError(f"Existing file: {saving_path}")

        # Create table columns
        cols = ["", ""]
        for model in self._loaded_results:
            cols.append(model)

        table_content = []

        if extration_profile_params is None and available_profile:
            extration_profile_params = {}

        # --------------------------------------------------------------------------------------------
        #                        Criteria integrated throughout the simulation
        # --------------------------------------------------------------------------------------------
        table_content.append(["Criteria integrated throughout the simulation"])

        # --------------------------------------- Flow area ------------------------------------------
        line = ["Flow Area", "Total Area [m2]"]
        _, area_num = self.compute_impacted_area(flow_threshold=flow_threshold)

        mean_list = []
        for model in self._loaded_results:
            mean_list.append(area_num[model])
            line.append(area_num[model])
        table_content.append(line)

        line = ["", "Compared to avrg"]
        mean_area = np.mean(np.array(mean_list))
        for model in self._loaded_results:
            line.append((area_num[model] - mean_area) / mean_area)
        table_content.append(line)

        line = ["Impacted Zone", "RMS (avrg)"]
        rms_area, _ = self.compute_impacted_area_rms_from_avrg(
            flow_threshold=flow_threshold
        )

        for model in self._loaded_results:
            line.append(rms_area[model])
        table_content.append(line)

        # ------------------------------------- Maximal height ---------------------------------------
        line = ["Maximal Height", "RMS (avrg)"]
        rms_height = self.compute_rms_from_avrg(output="h_max")

        for model in self._loaded_results:
            line.append(rms_height[model])
        table_content.append(line)

        # ------------------------------------- Maximal momentum ---------------------------------------
        line = ["Maximal Momentum", "RMS (avrg)"]
        rms_hu = self.compute_rms_from_avrg(output="hu_max")

        for model in self._loaded_results:
            line.append(rms_hu[model])
        table_content.append(line)

        # ----------------------------------- Average flow speed -------------------------------------
        if available_profile:
            avrg_vel, time, dist, model_pos = self.compute_average_velocity(
                distance=avrg_velocity_distance,
                look_up_direction=profile_direction,
                flow_threshold=flow_threshold,
                **extration_profile_params,
            )
            line = ["Velocity", f"Time [s] to complete d={dist}m"]

            for model in self._loaded_results:
                line.append(time[model])
            table_content.append(line)

            line = ["", "Average velocity [m/s]"]

            for model in self._loaded_results:
                line.append(avrg_vel[model])
            table_content.append(line)

        # --------------------------------------------------------------------------------------------
        #                       Criteria for the final time step of the simulation
        # --------------------------------------------------------------------------------------------
        table_content.append(["Criteria for the final time step of the simulation"])

        # --------------------------------------- Final height ---------------------------------------
        line = ["Final Thickness Repartition", "RMS (avrg)"]
        rms_height = self.compute_rms_from_avrg(output="h_final")

        for model in self._loaded_results:
            line.append(rms_height[model])
        table_content.append(line)

        # -------------------------------- Position of the flow front --------------------------------
        line = ["Maximal Extension", "Distance [m]"]
        _, model_max_dist = self.compute_dist_centermass(flow_threshold=flow_threshold)

        for model in self._loaded_results:
            line.append(model_max_dist[model][2])
        table_content.append(line)

        line = ["", "Compared to avrg"]

        list_pos = []
        for model in self._loaded_results:
            list_pos.append(model_max_dist[model][2])
        mean_dist = np.mean(np.array(list_pos))

        for model in self._loaded_results:
            line.append((model_max_dist[model][2] - mean_dist) / mean_dist)
        table_content.append(line)

        if available_profile:
            line = ["Flow Front (Profile)", "Distance [m]"]

            for model in self._loaded_results:
                line.append(model_pos[model][2])
            table_content.append(line)

            line = ["", "Compared to avrg"]

            list_pos = []
            for model in self._loaded_results:
                list_pos.append(model_pos[model][2])
            mean_dist = np.mean(np.array(list_pos))

            for model in self._loaded_results:
                line.append((model_pos[model][2] - mean_dist) / mean_dist)
            table_content.append(line)

        # ---------------------------------- Front shape comparison ----------------------------------
        if coussot_criteria is not None:
            line = ["Front Shape", "RMS (Coussot)"]
            model_rms, _, _ = self.compute_rms_from_coussot(
                look_up_direction=profile_direction,
                flow_threshold=flow_threshold,
                coussot_params=coussot_criteria,
            )

            for model in self._loaded_results:
                line.append(model_rms[model])
            table_content.append(line)

        # --------------------------------------------------------------------------------------------
        #                                     Numeralical criterias
        # --------------------------------------------------------------------------------------------
        table_content.append(["Numerical criteria"])

        # ------------------------------------------- Volume -----------------------------------------
        model_volume = {}
        for model in self._loaded_results:
            volume = self._loaded_results[model].get_output("volume")
            model_volume[model] = volume

        line = ["Volume Variation", "RMS of Total Volume Variation"]

        for model in self._loaded_results:
            rms_vinit = 0
            for v in model_volume[model].d:
                rms_vinit += (
                    np.sqrt((v - model_volume[model].d[0]) ** 2)
                    / model_volume[model].d[0]
                )
            line.append(rms_vinit)
        table_content.append(line)

        line = ["", "(Vf - Vi) / Vi"]

        for model in self._loaded_results:
            line.append(
                (model_volume[model].d[-1] - model_volume[model].d[0])
                / model_volume[model].d[0]
            )
        table_content.append(line)

        # --------------------------------------------------------------------------------------------
        #                                          Saving table
        # --------------------------------------------------------------------------------------------
        df = pd.DataFrame(table_content, columns=cols)
        if save:
            if fmt == "csv":
                df.to_csv(saving_path, index=False, encoding="utf-8")

            if fmt == "xlsx":
                df.to_excel(saving_path, index=False, engine="openpyxl")

                # --------------------------------------- Reformat table -------------------------------------
                wb = openpyxl.load_workbook(saving_path)
                ws = wb.active

                # Add border style
                bordure_fine = openpyxl.styles.Border(
                    left=openpyxl.styles.Side(style="thin"),
                    right=openpyxl.styles.Side(style="thin"),
                    top=openpyxl.styles.Side(style="thin"),
                    bottom=openpyxl.styles.Side(style="thin"),
                )

                for row in ws.iter_rows(
                    min_row=1, max_row=ws.max_row, max_col=ws.max_column
                ):
                    for cell in row:
                        cell.border = bordure_fine
                wb.save(saving_path)

            print(f"Saved in: {saving_path}")

        return df

    def generate_analytical_comparison_csv(
        self,
        analytic_solution: dict,
        save: bool = False,
        folder_out: str = None,
        file_name: str = None,
        fmt: str = "csv",
        compute_as_u: bool = True,
        extration_profile_params: dict = None,
        flow_threshold: float = None,
    ) -> None:
        """Generate a csv file summarizing comparison criteria between flow models and analytic solution.

        Generates a file grouping comparison criteria between numerical flow models and analytic solution:

            - Criteria integrated throughout the simulation:

                - Total Height: RMS of flow height versus analytical solution.
                - Total Momentum: RMS of flow h*u versus analytical solution.
                - Total Front Position: RMS of flow front versus analytical solution.

            - Criteria for the final time step of the simulation:

                - Final Height: RMS of flow final height versus analytical solution.
                - Flow Front Position: maximum distance traveled by the flow and comparison with the analytic solution.

            - Numerical criteria:

                - Volume: value of the volume at final time steps (compared to initial volume) and RMS versus initial volume value.

        Parameters
        ----------
        analytic_solution: dict
            Argument for the analytic solution. See :meth:`compute_analytic_solution`.
        save : bool, optional
            If True, save the resulting tab at :data:`folder_out`. By default False.
        folder_out : str, optional
            Path to the folder where the file is saved, if None generate "xlsx_results"
            folder in code folder. By default None
        file_name : str, optional
            Name of the folder, if None use "AS_comparison_[models]". By default None.
        fmt : str, optional
            Saving format of the table. Can be "csv" or "xlsx". By default "csv".
        compute_as_u: bool, optional
            If True, compute analytic solution for flow velocity. Can be disabled. By default True.
        extration_profile_params: dict, optional
            Argument for extracting profile. See :meth:`tilupy.read.Results.get_profile`. By default None.
        flow_threshold : float, optional
            Flow threshold when extracting front position from profile, by default None.

        Raises
        ------
        ValueError
            If the file already exists.
        """
        import pandas as pd
        import openpyxl
        import os

        if save:
            if folder_out is None:
                folder_out = os.path.join(os.path.dirname(__file__), "xlsx_results")

            os.makedirs(folder_out, exist_ok=True)

            if file_name is None:
                file_name = "AS_compare"
                for model in self._loaded_results:
                    file_name += "_" + model

            if not file_name.endswith("." + fmt):
                file_name = file_name + "." + fmt
            saving_path = os.path.join(folder_out, file_name)

            if os.path.exists(saving_path):
                raise ValueError(f"Existing file: {saving_path}")

        extration_profile_params = (
            {} if extration_profile_params is None else extration_profile_params
        )

        # Create table columns
        cols = ["", ""]
        for model in self._loaded_results:
            cols.append(model)

        table_content = []

        # --------------------------------------------------------------------------------------------
        #                        Compute AS at each time steps for each models
        # --------------------------------------------------------------------------------------------
        model_AS_h = {}
        model_AS_u = {}
        for model in self._loaded_results:
            as_h_profile = self.compute_analytic_solution(
                output="h", T=self._models_tim[model], model=model, **analytic_solution
            )
            model_AS_h[model] = as_h_profile
            if compute_as_u:
                as_u_profile = self.compute_analytic_solution(
                    output="u",
                    T=self._models_tim[model],
                    model=model,
                    **analytic_solution,
                )

                model_AS_u[model] = np.nan_to_num(as_u_profile.d, nan=0)

            if flow_threshold is None:
                flow_threshold = np.max(model_AS_h[model].d) * 0.01

        # --------------------------------------------------------------------------------------------
        #                        Criteria integrated throughout the simulation
        # --------------------------------------------------------------------------------------------
        table_content.append(["Criteria integrated throughout the simulation"])

        # ---------------------------------- Height difference ---------------------------------------
        model_h = {}
        for model in self._loaded_results:
            profile, _ = self._loaded_results[model].get_profile(
                output="h", **extration_profile_params
            )
            model_h[model] = profile

        line = ["Total Height Difference", "RMS (AS)"]
        for model in self._loaded_results:
            line.append(
                (
                    np.sqrt(np.sum((model_h[model].d - model_AS_h[model].d) ** 2))
                    / np.sqrt(np.sum((model_AS_h[model].d) ** 2))
                )
            )

        table_content.append(line)

        # ------------------------------------ HU difference -----------------------------------------
        if compute_as_u:
            model_u = {}
            for model in self._loaded_results:
                profile, _ = self._loaded_results[model].get_profile(
                    output="u", **extration_profile_params
                )
                model_u[model] = profile

            line = ["Total Momentum Difference", "RMS (AS)"]
            for model in self._loaded_results:
                hu = model_h[model].d * model_u[model].d
                as_hu = model_AS_h[model].d * model_AS_u[model]
                line.append(
                    (np.sqrt(np.sum((hu - as_hu) ** 2)) / np.sqrt(np.sum((as_hu) ** 2)))
                )

            table_content.append(line)

        # ----------------------------------- Front position ---------------------------------------
        model_position = {}
        model_as_position = {}
        for model in self._loaded_results:
            list_position = []
            list_as_position = []
            for t in range(len(model_h[model].t)):
                # Model
                max_index = np.argmax(model_h[model].d[:, t])
                idx = np.where(model_h[model].d[max_index:, t] <= flow_threshold)[0]
                idx = (
                    (idx[0] - 1)
                    if len(idx)
                    else len(model_h[model].d[max_index:, t]) - 1
                )
                idx += max_index
                list_position.append(model_h[model].coords[idx])
                # AS
                max_index = np.argmax(model_AS_h[model].d[:, t])
                idx = np.where(model_AS_h[model].d[max_index:, t] <= flow_threshold)[0]
                idx = (
                    (idx[0] - 1)
                    if len(idx)
                    else len(model_AS_h[model].d[max_index:, t]) - 1
                )
                idx += max_index
                list_as_position.append(model_AS_h[model].coords[idx])
            model_position[model] = np.array(list_position)
            model_as_position[model] = np.array(list_as_position)

        line = ["Total Front Position Difference", "RMS (AS)"]

        for model in self._loaded_results:
            line.append(
                (
                    np.sqrt(
                        np.sum((model_position[model] - model_as_position[model]) ** 2)
                    )
                    / np.sqrt(np.sum((model_as_position[model]) ** 2))
                )
            )

        table_content.append(line)

        # --------------------------------------------------------------------------------------------
        #                       Criteria for the final time step of the simulation
        # --------------------------------------------------------------------------------------------
        table_content.append(["Criteria for the final time step of the simulation"])

        # --------------------------------------- Final height ---------------------------------------
        line = ["Final Thickness Repartition", "RMS (AS)"]
        for model in self._loaded_results:
            line.append(
                (
                    np.sqrt(
                        np.sum(
                            (model_h[model].d[:, -1] - model_AS_h[model].d[:, -1]) ** 2
                        )
                    )
                    / np.sqrt(np.sum((model_AS_h[model].d[:, -1]) ** 2))
                )
            )

        table_content.append(line)

        # ---------------------------- Front position at final time step -----------------------------
        line = ["Final Front Position", "Distance [m]"]

        for model in self._loaded_results:
            line.append(model_position[model][-1] - model_position[model][0])

        table_content.append(line)

        line = ["", "Compared to AS"]

        # model_as_position[model] différent selon model
        for model in self._loaded_results:
            line.append(
                (model_position[model][-1] - model_as_position[model][-1])
                / model_as_position[model][-1]
            )

        table_content.append(line)

        # --------------------------------------------------------------------------------------------
        #                                     Numeralical criteria
        # --------------------------------------------------------------------------------------------
        table_content.append(["Numerical criteria"])

        # ------------------------------------------- Volume -----------------------------------------
        model_volume = {}
        for model in self._loaded_results:
            volume = self._loaded_results[model].get_output("volume")
            model_volume[model] = volume

        line = ["Volume Variation", "RMS of Total Volume Variation"]

        for model in self._loaded_results:
            rms_vinit = 0
            for v in model_volume[model].d:
                rms_vinit += (
                    np.sqrt((v - model_volume[model].d[0]) ** 2)
                    / model_volume[model].d[0]
                )
            line.append(rms_vinit)
        table_content.append(line)

        line = ["", "(Vf - Vi) / Vi"]

        for model in self._loaded_results:
            line.append(
                (model_volume[model].d[-1] - model_volume[model].d[0])
                / model_volume[model].d[0]
            )
        table_content.append(line)

        # --------------------------------------------------------------------------------------------
        #                                          Saving table
        # --------------------------------------------------------------------------------------------
        df = pd.DataFrame(table_content, columns=cols)
        if save:
            if fmt == "csv":
                df.to_csv(saving_path, index=False, encoding="utf-8")

            if fmt == "xlsx":
                df.to_excel(saving_path, index=False, engine="openpyxl")

                # --------------------------------------- Reformat table -------------------------------------
                wb = openpyxl.load_workbook(saving_path)
                ws = wb.active

                # Add border style
                bordure_fine = openpyxl.styles.Border(
                    left=openpyxl.styles.Side(style="thin"),
                    right=openpyxl.styles.Side(style="thin"),
                    top=openpyxl.styles.Side(style="thin"),
                    bottom=openpyxl.styles.Side(style="thin"),
                )

                for row in ws.iter_rows(
                    min_row=1, max_row=ws.max_row, max_col=ws.max_column
                ):
                    for cell in row:
                        cell.border = bordure_fine
                wb.save(saving_path)

            print(f"Saved in: {saving_path}")

        return df



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
import importlib


def read_raster(file: str) -> list[np.ndarray]:
    """Convert a raster file (tif or asc) into numpy array.

    Parameters
    ----------
    file : str
        Path to the raster file.

    Returns
    -------
    list[np.ndarray]
        X and Y coordinates and data values in numpy array.
    """
    if file.endswith(".asc") or file.endswith(".txt"):
        return read_ascii(file)
    elif file.endswith(".tif") or file.endswith(".tif"):
        return read_tiff(file)


def read_tiff(file: str) -> list[np.ndarray]:
    """Read and convert a tiff file into numpy array.

    Parameters
    ----------
    file : str
        Path to the tiff file.

    Returns
    -------
    list[np.ndarray]
        X and Y coordinates and data values in numpy array.
    """
    import rasterio

    with rasterio.open(file, "r") as src:
        dem = src.read(1)
        ny, nx = dem.shape
        x = np.linspace(src.bounds.left, src.bounds.right, nx)
        y = np.linspace(src.bounds.bottom, src.bounds.top, ny)
    return x, y, dem


def read_ascii(file: str) -> list[np.ndarray]:
    """Read and convert a ascii file into numpy array.

    Parameters
    ----------
    file : str
        Path to the ascii file.

    Returns
    -------
    list[np.ndarray]
        X and Y coordinates and data values in numpy array.
    """
    dem = np.loadtxt(file, skiprows=6)
    grid = {}
    with open(file, "r") as fid:
        for i in range(6):
            tmp = fid.readline().split()
            grid[tmp[0]] = float(tmp[1])
    try:
        x0 = grid["xllcenter"]
        y0 = grid["yllcenter"]
    except KeyError:
        x0 = grid["xllcorner"]
        y0 = grid["yllcorner"]
    nx = int(grid["ncols"])
    ny = int(grid["nrows"])
    dx = dy = grid["cellsize"]
    x = np.linspace(x0, x0 + (nx - 1) * dx, nx)
    y = np.linspace(y0, y0 + (ny - 1) * dy, ny)

    return x, y, dem


def write_tiff(
    x: np.ndarray, y: np.ndarray, z: np.ndarray, file_out: str, **kwargs
) -> None:
    """Write tif file from numpy array.

    Parameters
    ----------
    x : np.ndarray
        X coordinates.
    y : np.ndarray
        Y coordinates.
    z : np.ndarray
        Elevation values.
    file_out : str
        Name of the output folder.
    """
    import rasterio
    from rasterio.transform import Affine

    if "driver" not in kwargs:
        kwargs["driver"] = "GTiff"
    res = (x[-1] - x[0]) / (len(x) - 1)
    transform = Affine.translation(x[0] - res / 2, y[-1] - res / 2) * Affine.scale(
        res, -res
    )

    with rasterio.open(
        file_out,
        "w",
        height=z.shape[0],
        width=z.shape[1],
        count=1,
        dtype=z.dtype,
        transform=transform,
        **kwargs,
    ) as dst:
        dst.write(z, 1)


def write_ascii(
    x: np.ndarray,
    y: np.ndarray,
    z: np.ndarray,
    file_out: str,
) -> None:
    """Write ascii file from numpy array.

    Parameters
    ----------
    x : np.ndarray
        X coordinates.
    y : np.ndarray
        Y coordinates.
    z : np.ndarray
        Elevation values.
    file_out : str
        Name of the output folder.
    """
    nx = z.shape[1]
    ny = z.shape[0]
    cellsize = x[1] - x[0]
    header_txt = "ncols {:.0f}\nnrows {:.0f}\nxllcorner {:.5f}\nyllcorner {:.5f}\n"
    header_txt += "cellsize {:.4f}\nnodata_value -99999"
    header_txt = header_txt.format(nx, ny, x[0], y[0], cellsize)
    np.savetxt(file_out, z, header=header_txt, comments="")


def write_raster(
    x: np.ndarray,
    y: np.ndarray,
    z: np.ndarray,
    file_out: str,
    fmt: str = None,
    **kwargs,
) -> None:
    """Write raster file from numpy array.

    Parameters
    ----------
    x : np.ndarray
        X coordinates.
    y : np.ndarray
        Y coordinates.
    z : np.ndarray
        Elevation values.
    file_out : str
        Name of the output folder.
    fmt : str
        Wanted format : "asc", "ascii", "txt", "tif", "tiff".

    Raises
    ------
    ValueError
        If invalid format.
    """
    # File format read from file_out overrides fmt
    fmt_tmp = file_out.split(".")
    if len(fmt_tmp) > 1:
        fmt = fmt_tmp[-1]
    else:
        if fmt is None:
            fmt = "asc"
        file_out = file_out + "." + fmt

    if fmt not in ["asc", "ascii", "txt", "tif", "tiff"]:
        raise ValueError("File format not implemented in write_raster")

    if fmt.startswith("tif"):
        if importlib.util.find_spec("rasterio") is None:
            print(
                (
                    "rasterio is required to write tif files.",
                    " Switching to asc format",
                )
            )
            fmt = "asc"

    if fmt in ["asc", "ascii", "txt"]:
        write_ascii(x, y, z, file_out)
    elif fmt in ["tif", "tiff"]:
        write_tiff(x, y, z, file_out, **kwargs)
    else:
        raise NotImplementedError()



# -*- coding: utf-8 -*-

import requests
import zipfile
import io
import os
import pandas as pd


def import_frankslide_dem(folder_out: str = None, file_out: str = None) -> str:
    """Import frankslide topography.

    Parameters
    ----------
    folder_out : str, optional
        Path to the folder output. If None the current folder will be choosed. By default None.
    file_out : str, optional
        Name of the file. If None choose "Frankslide_topography.asc". By default None.

    Returns
    -------
    str
        Path to the saved file.
    """
    if folder_out is None:
        folder_out = "."
    if file_out is None:
        file_out = "Frankslide_topography.asc"

    file_save = os.path.join(folder_out, file_out)

    url = (
        "https://raw.githubusercontent.com/marcperuz/tilupy/main/data/"
        + "frankslide/rasters/Frankslide_topography.asc"
    )
    r = requests.get(url)
    open(file_save, "w").write(r.text)

    return file_save


def import_frankslide_pile(folder_out: str = None, file_out: str = None) -> str:
    """Import frankslide pile.

    Parameters
    ----------
    folder_out : str, optional
        Path to the folder output. If None the current folder will be choosed. By default None.
    file_out : str, optional
        Name of the file. If None choose "Frankslide_pile.asc". By default None.

    Returns
    -------
    str
        Path to the saved file.
    """

    if folder_out is None:
        folder_out = "."
    if file_out is None:
        file_out = "Frankslide_pile.asc"

    file_save = os.path.join(folder_out, file_out)

    url = (
        "https://raw.githubusercontent.com/marcperuz/tilupy/main/data/"
        + "frankslide/rasters/Frankslide_pile.asc"
    )
    r = requests.get(url)
    open(file_save, "w").write(r.text)

    return file_save


def import_shaltop_frankslide(folder_out: str = "./shaltop_frankslide"):
    """Import shaltop results for the Frankslide.

    Parameters
    ----------
    folder_out : str, optional
        Folder where data will be saved. By default "./shaltop_frankslide".

    Returns
    -------
    None

    """
    url = (
        "https://raw.githubusercontent.com/marcperuz/tilupy/"
        + "main/data/shaltop/shaltop_frankslide.zip"
    )
    r = requests.get(url)
    z = zipfile.ZipFile(io.BytesIO(r.content))
    z.extractall(folder_out)


def import_shaltop_mus_calibrated(sep=";"):
    """Import database of friction coefficients calibrated with Shaltop with the Coulomb rheology"""

    # Load data
    url = "https://zenodo.org/records/18791119/files/mus_calibrated.csv"
    response = requests.get(url, stream=True)
    response.raise_for_status()
    contenu = io.StringIO(response.text)
    df = pd.read_csv(contenu, sep=sep)

    # Get publication date
    params = {
        "sort": "mostrecent",
        "page": 1,
        "size": 1,  # On veut juste la plus récente
        "q": "conceptrecid:18791118",
    }
    response = requests.get("https://zenodo.org/api/records", params=params)
    if response.status_code != 200:
        raise Exception(f"Request error : {response.status_code}, {response.text}")
    data = response.json()
    if data["hits"]["hits"]:
        latest = data["hits"]["hits"][0]
        publication_date = latest["metadata"]["publication_date"]
    else:
        raise Exception("conceptrecid 18791118 not found.")

    return df, publication_date



# -*- coding: utf-8 -*-

import os

import numpy as np
import tilupy.read

# Classify results
AVAILABLE_OUTPUT = ["h", "hvert", "u", "hu", "hu2", "ux", "uy", "hux", "huy"]
"""All output available for a shaltop simulation.

Implemented states:

    - h : Flow thickness (normal to the surface)
    - hux : Momentum flux component in X direction
    - huy : Momentum flux component in Y direction
    
Output computed from other output:
    
    - hvert : True vertical flow thickness
    - ux : Velocity component in X direction
    - uy : Velocity component in Y direction
    - u : Norm of the velocity (from ux and uy)
    - hu : Momentum flux (from h and u)
    - hu2 : Convective momentum flux (from h and u)
"""


def extract_saval_ascii(path: str) -> np.ndarray:
    """Extract ascii file data generated by saval2D.

    Parameters
    ----------
    path : str
        Path to the ascii file.

    Returns
    -------
    numpy.ndarray
        Values in the ascii file.
    """
    with open(path, "r") as f:
        lines = f.readlines()
    start_index = 6

    h_str = lines[start_index:]

    h = [list(map(float, line.split())) for line in h_str if line.strip()]
    return np.array(h)


class Results(tilupy.read.Results):
    """Results of saval2D simulations.

    This class is the results class for saval2D. Reading results from saval2D outputs
    are done in this class.

    This class has all the global and quick attributes of the parent class. The quick
    attributes are only computed if needed and can be deleted to clean memory.

    In addition to these attributes, there are those necessary for the operation of
    reading the saval2D results.

    Parameters
    ----------
    folder : str
        Path to the folder containing the simulation files.
    raster_topo : str
        Topographic raster name.

    Attributes
    ----------
        _simu_inputs : str
            Path to the input folder of saval2D simulation.
        _simu_outputs : str
            Path to the output folder of saval2D simulation.
        _raster : str
            Raster file name.
        _params : dict
            Dictionary storing all simulation parameters.
    """

    def __init__(self, folder, raster_topo="mntsimulation.asc"):
        super().__init__()
        self._code = "saval2D"

        self._folder = folder
        self._folder_output = folder

        self._simu_inputs = (
            os.path.join(folder, "inputs")
            if os.path.exists(os.path.join(folder, "inputs"))
            else None
        )
        self._simu_outputs = (
            os.path.join(folder, "outputs")
            if os.path.exists(os.path.join(folder, "outputs"))
            else None
        )

        if not raster_topo.endswith(".asc"):
            raster_topo = raster_topo + ".asc"
        self._raster = raster_topo

        self._params = dict()

        if self._simu_inputs is not None:
            self._x, self._y, self._zinit = tilupy.raster.read_ascii(
                os.path.join(self._simu_inputs, self._raster)
            )
        self._nx, self._ny = len(self._x), len(self._y)
        self._dx = self._x[1] - self._x[0]
        self._dy = self._y[1] - self._y[0]

        self._params["nx"] = len(self.x)
        self._params["ny"] = len(self.y)

        # Rheology
        rheol = self.get_rheological_parameters(
            os.path.join(self._simu_outputs, "log.txt")
        )
        self._params["tau/rho"] = rheol[0]
        self._params["mu"] = rheol[1]
        self._params["xi"] = rheol[2]

        # Create self._tim
        self._extract_output("X")

    def get_rheological_parameters(self, path_to_log: str) -> list[str]:
        """Get rheological parameters of a saval2D simulation.

        Parameters
        ----------
        path_to_log : str
            Path to the log file containing the value of the rheological parameters.

        Returns
        -------
        list[str]
            List of rheological parameters : tau/rho, mu and xi.
        """
        values = None
        with open(path_to_log, "r") as f:
            for line in f:
                line = line.strip()
                try:
                    parts = list(map(float, line.split()))
                    if len(parts) == 3:
                        values = parts
                        break
                except ValueError:
                    continue
        return values

    def get_times(self, path: str) -> list[float]:
        """Get simulation time steps.

        Parameters
        ----------
        path : str
            Path to the log file.

        Returns
        -------
        list[float]
            List of recorded time steps.
        """
        t_list = []
        with open(path, "r") as f:
            lines = f.readlines()

        start_index = 16
        for line in lines[start_index:]:
            if line.strip() == "":
                continue
            else:
                _, t, _, _, _ = line.split()
                t_list.append(float(t))

        return t_list

    def _extract_output(
        self, name: str, **kwargs
    ) -> tilupy.read.TemporalResults2D | tilupy.read.AbstractResults:
        """Result extraction for saval2D files.

        Parameters
        ----------
        name : str
            Wanted output. Can access to variables in :data:`AVAILABLE_OUTPUT`.

        Returns
        -------
        tilupy.read.TemporalResults2D | tilupy.read.AbstractResults
            Wanted output. If no output computed, return an object of :class:`tilupy.read.AbstractResults`.
        """
        # Read thicknesses or velocity components
        d = None
        t = None
        notation = None

        tim = [0]
        t_list = self.get_times(os.path.join(self._simu_outputs, "log.txt"))

        _, _, h_init = tilupy.raster.read_ascii(
            os.path.join(self._simu_inputs, "zdepsimulation.asc")
        )

        h_list = [h_init]

        ux_list = []
        uy_list = []
        u_list = []
        ux_list.append(np.zeros_like(h_init))
        uy_list.append(np.zeros_like(h_init))
        u_list.append(np.zeros_like(h_init))

        qx_list = []
        qy_list = []
        q_list = []
        qx_list.append(np.zeros_like(h_init))
        qy_list.append(np.zeros_like(h_init))
        q_list.append(np.zeros_like(h_init))

        for T in range(len(t_list)):
            h_t = extract_saval_ascii(
                os.path.join(self._simu_outputs, f"resuh{T + 1}.asc")
            )
            qu_t = extract_saval_ascii(
                os.path.join(self._simu_outputs, f"resuqu{T + 1}.asc")
            )
            qv_t = extract_saval_ascii(
                os.path.join(self._simu_outputs, f"resuqv{T + 1}.asc")
            )

            # Condition to avoid flow velocity on part with no fluid
            h_t[h_t < 0.0001] = 0

            ux_t = np.divide(qu_t, h_t, out=np.zeros_like(qu_t), where=h_t != 0)
            uy_t = np.divide(qv_t, h_t, out=np.zeros_like(qv_t), where=h_t != 0)

            u_t = np.sqrt(ux_t**2 + uy_t**2)
            q_t = np.sqrt(qu_t**2 + qv_t**2)

            h_list.append(h_t)

            ux_list.append(ux_t)
            uy_list.append(uy_t)
            u_list.append(u_t)

            qx_list.append(qu_t)
            qy_list.append(qv_t)
            q_list.append(q_t)

            tim.append(t_list[T])

        if self._tim is None:
            self._tim = np.array(tim)

        extracted_outputs = {
            "h": np.stack(h_list, axis=-1),
            "u": np.stack(u_list, axis=-1),
            "hu": np.stack(q_list, axis=-1),
            "ux": np.stack(ux_list, axis=-1),
            "uy": np.stack(uy_list, axis=-1),
            "hux": np.stack(qx_list, axis=-1),
            "huy": np.stack(qy_list, axis=-1),
        }

        if name in ["h", "u", "hu", "ux", "uy", "hux", "huy"]:
            d = extracted_outputs[name]
            t = self._tim

        if name == "hu2":
            d = extracted_outputs["hu"] * extracted_outputs["u"]
            t = self._tim

        elif name == "hvert":
            if self._costh is None:
                self._costh = self.compute_costh()
            d = extracted_outputs["h"] / self._costh[:, :, np.newaxis]
            t = self._tim

        # if name == "ek":
        #     if self._costh is None:
        #             self._costh = self.compute_costh()

        #     d = []
        #     for i in range(len(self._tim)):
        #         d.append(np.sum((np.stack(available_outputs["hu"], axis=-1)[:, :, i] *
        #                          np.stack(available_outputs["u"], axis=-1)[:, :, i] *
        #                          self._dx *
        #                          self._dy)
        #                         / self._costh[:, :]))
        #     d = np.array(d)
        #     t = self._tim

        if t is None:
            return tilupy.read.AbstractResults(name, d, notation=notation)

        else:
            if d.ndim == 3:
                return tilupy.read.TemporalResults2D(
                    name, d, t, notation=notation, x=self._x, y=self._y, z=self._zinit
                )
        return None

    def _read_from_file(self, *args, **kwargs):
        """Not useful"""
        return "No _read_from_file for Saval2D."
